﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=10
@EndOfDesignText@
#Region Code Module Information
'Static Code Module: Helper
'File: Helper.bas
'Static code module with generic helper methods used in all modules.
'
'PFDCreator Copyright see B4XMainPage.
#End Region

Sub Process_Globals
	'Core
	Private xui As XUI
	Private fx As JFX

	'Messages
	Public MSG_TYPE_INFORMATION As Int = 1
	Public MSG_TYPE_WARNING As Int = 2
	Public MSG_TYPE_ERROR As Int = 3
End Sub

#Region Logging
'Write to the B4j IDE is flag debug is set.
Public Sub ToLog(debug As Boolean, text As String)
	If debug And text.Length > 0 Then
		Log(text)
	End If
End Sub
#End Region

#Region DateTime
'Get the current date and time.
Public Sub GetDateTime As String
	Return $"${DateTime.Date(DateTime.Now)} ${DateTime.Time(DateTime.now)}"$
End Sub
#End Region

#Region Numbers
'Map the value (int) from range low - high to new range low - high.
'Example: MapRange(50, 0, 100, 0, 255) = 127
Public Sub MapRange(value As Int, fromlow As Int, fromhigh As Int, tolow As Int, tohigh As Int) As Int
	Return tolow + (tohigh - tolow) * ((value - fromlow) / (fromhigh - fromlow))
End Sub

'Map the value (float) from range low - high to new range low - high.
'Example: MapRange(50, 0, 100, 0, 255) = 127
Public Sub MapRangeFloat(value As Float, fromlow As Float, fromhigh As Float, tolow As Float, tohigh As Float) As Float
	Return tolow + (tohigh - tolow) * ((value - fromlow) / (fromhigh - fromlow))
End Sub

'Check if a value is between a min/max range.
'Example: value=10, rangemin=8, rangemax=12 result is true. For value=7 result is false.
Public Sub InRange(value As Int, rangemin As Int, rangemax As Int) As Boolean
	Dim result As Boolean = False
	
	If value >= rangemin And value <= rangemax Then
		result = True
	End If

	' ToLog(True, $"[InRange]result=${result},rangemin=${rangemin},value=${value},rangemax=${rangemax}"$)
	Return result
End Sub

'Check if a string is a number. Return true is number.
Public Sub ValueIsNumber(s As String) As Boolean
	Return IsNumber(s) And Regex.IsMatch("[\d\.]+", s)
End Sub

'Check if a string is a number.
'If false, 0 is returned else the value as double.
Public Sub CheckNumberField(value As String) As Double
	If ValueIsNumber(value) Then
		Return value.As(Double)
	Else
		Log($"[ERROR][CheckNumberField] ${value} is not a number."$)
		Return 0
	End If
End Sub

'Return a number with no decimals and remove comma (), or dot (.).
Public Sub SetNumberFormatNoDecimal(value As Double) As String
	Dim n As String = NumberFormat(value,0,0)
	n = n.Replace(",", "")
	n = n.Replace(".", "")
	Return n
End Sub

'Convert percentage to integer.
'Used to calculate for example the width.
Public Sub PerToCurrent(per As Int, current As Int) As Int
	Return Round(current * (per * 0.01))
End Sub
#End Region

#Region Font
'Create a B4X font italic with font size.
'The font is converted from FX font to B4X font.
Public Sub CreateDefaultItalicFont(fontsize As Int) As B4XFont
	'Get the familyname of the default font
	Dim fontfamily As String = GetFontDefaultFamilyName
	'Create fx font italic
	Dim fntfx As Font = fx.CreateFont(fontfamily, fontsize, False, True)
	'Convert the fx font italic to B4XFont
	Return xui.CreateFont(fntfx, fontsize)
End Sub

'Get the list of default font names used.
'Supported are only specific fonts because these are created.
'Returns (List) - List with default font names, like DEFAULT, DEFAULTBOLD
Public Sub GetDefaultFontNames As List
	Dim result As List
	result.Initialize
	'Add the default specific from PFDCommon string array PFDFonts()
	For Each s As String In PFDCommon.PFDFonts
		result.InsertAt(0,s)		
	Next
	'NOT USED - but manual option to insert
	'result.InsertAt(0,"DIGITAL")
	'result.InsertAt(0,"DEFAULTBOLD")
	'result.InsertAt(0,"DEFAULTITALIC")
	'result.InsertAt(0,"DEFAULT")
	Return result
End Sub

'Get the list of font names installed.
'Returns (List) - List with font family names, like Arial,Arial Black,Arial Narrow,...
Public Sub GetFontNames As List
	Dim result As List
	result.Initialize
	For Each familyname As String In fx.GetAllFontFamilies
		result.Add(familyname)
		'Log(familyname)
	Next
	'Add the default specific from PFDCommon string array PFDFonts()
	For Each s As String In PFDCommon.PFDFonts
		result.InsertAt(0,s)
	Next
	Return result
End Sub

'Get the default B4XFont family name by creating a xui default font.
Public Sub GetFontDefaultFamilyName As String
	Dim fnt As B4XFont = xui.CreateDefaultFont(14)
	Return fnt.ToNativeFont.FamilyName
End Sub
#End Region

#Region Strings
'Check if a string item is in array of strings.
'Return True if found else False
'Example<code>
'If Not(ItemInArray(obj.direction, Array As String(PFDCommon.DIRECTION_HORIZONTAL, PFDCommon.DIRECTION_VERTICAL))) Then
'	obj.direction = PFDCommon.DIRECTION_HORIZONTAL
'End If
'</code>
Public Sub ItemInArray(item As String, items() As String) As Boolean
	Dim result As Boolean = False
	For Each s As String In items
		If s == item Then Return True
	Next
	Return result
End Sub

'Count the occurence of a character in a string
Public Sub StringCountChar(str As String, ch As Char) As Int
	Dim result As Int = 0
	Dim strarr() As String = Regex.Split("", str)
	For Each s As String In strarr
		If s.EqualsIgnoreCase(ch) Then result = result + 1
	Next
	Return result
End Sub

Public Sub StringReplaceFirstChar(str As String, oldch As Char, newch As Char) As String
	Dim result As String = ""
	Dim replaced As Boolean = False
	Dim sb As StringBuilder
	sb.Initialize
	
	Dim index As Int = str.IndexOf(oldch)
	If index > -1 Then
		Dim strarr() As String = Regex.Split("", str)
		For Each s As String In strarr
			If s.EqualsIgnoreCase(oldch) And Not(replaced) Then
				s = newch
				replaced = True
			End If
			sb.Append(s)
		Next
		result = sb.ToString
	Else
		result = str
	End If
	Return result
End Sub

'Wrapping a string into multiline string.
'text - String to wrap
'n - Max number of characters on a line
'Returns wrapped text with CRLF
Public Sub StringWrap(text As String, n As Int) As String
	Dim result As String = text
	Dim index As Int = 0
	
	Do While result.Length > index + n
		Dim lastIndex As Int = result.LastIndexOf2(" ", index + n)
		If lastIndex = -1 Then
			'LogError("[TextWrap2] Impossible to slice: " & result.SubString2(0, n))
			Return ""
		End If
		result = result.SubString2(0, lastIndex) & CRLF & result.SubString(lastIndex + 1)
		index = lastIndex + 1
	Loop
	
	Return result
End Sub

'Wrapping a string into multiline string.
'Returns a map with the text and linecount.
'text - String to wrap
'n - Max number of characters on a line
'Returns Map with the wrapped text and the number of lines.
'text = m.get("text"), linecount = m.get("cnt")
Public Sub StringWrap2(text As String, n As Int) As Map
	Dim result As String = text
	Dim index As Int = 0
	Dim linecnt As Int = 0
	
	Do While result.Length > index + n
		Dim lastIndex As Int = result.LastIndexOf2(" ", index + n)
		If lastIndex = -1 Then
			'LogError("[TextWrap2] Impossible to slice: " & result.SubString2(0, n))
			Return CreateMap("text":text,"cnt":1)
		End If
		result = result.SubString2(0, lastIndex) & CRLF & result.SubString(lastIndex + 1)
		linecnt = linecnt + 1
		index = lastIndex + 1
	Loop
	If result.LastIndexOf(CRLF) == result.Length Then
		result = result.SubString2(0, result.Length - 1)
		linecnt = linecnt - 1
	Else
		linecnt = linecnt + 1
	End If
	Return CreateMap("text":result,"cnt":linecnt)
End Sub

'Word wrap the object value by the object width to a list of lines.
'Example<code>
'Dim WrappedLines As List = WordWrapString(mCvs, obj)
'Dim fnt As B4XFont = PFDCommon.SetFont(obj)
'Dim fntrect As B4XRect = cvs.MeasureText("A", fnt)
'Dim x, y As Float
'y = obj.centery - ( (fntrect.Height * WrappedLines.Size) / 2)    
'For Each Line As String In WrappedLines
'	x = obj.centerx
'	y = y + fntrect.Height + 2dip
'	cvs.DrawText(Line, x, y, fnt, obj.textcolor, obj.textalignment)
'Next
'</code>
Public Sub WordWrapString(cvs As AsyncCanvas, obj As TPFDObject) As List
   	Dim Lines As List
	Dim Words() As String = Regex.Split(" ", obj.value)
	Dim CurrentLine As StringBuilder
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)

	'Init the lines being returned
	Lines.Initialize
	'Init the current line holding the words seperated by space
	CurrentLine.Initialize

	'Loop over the words, check if not empty and add to the current line or add the current line to the lines
	For Each Word As String In Words
		' Add the word to the current line and measure its width
		Dim TempLine As String = CurrentLine.ToString & (IIf(CurrentLine.Length > 0, " ", "")) & Word
		Dim fntrect As B4XRect = cvs.MeasureText(TempLine, fnt)
		Dim LineWidth As Float = fntrect.width
        
		If LineWidth > obj.Width Then
			If Word.Length > 0 Then
				'If the line exceeds obj.width, add the current line to Lines and start a new one
				If CurrentLine.ToString.Length > 0 Then Lines.Add(CurrentLine.ToString)
				CurrentLine.Initialize
				CurrentLine.Append(Word)
			End If
		Else
			If Word.Length > 0 Then
				'Otherwise, keep adding the word to the current line
				If CurrentLine.Length > 0 Then CurrentLine.Append(" ")
				CurrentLine.Append(Word)
			End If
		End If
	Next
    
	'Add the last line
	If CurrentLine.Length > 0 Then Lines.Add(CurrentLine.ToString)

	'Return the list of lines
	Return Lines
End Sub

'Wrapping a text into a b4xbitmap.
'Create a b4xbitmap with wrapped text from the obj.value using a label.
'Example<code>
'Dim bmp As B4XBitmap = Helper.TextWrappedToBitMap(cvs, obj)
'Dim rect As B4XRect
'rect.Initialize(obj.centerx - (obj.width/2),obj.centery - (obj.height/2),obj.centerx + (obj.width/2),obj.centery + (obj.height/2))
'cvs.DrawBitmap(bmp, rect)
'</code>
Public Sub TextWrappedToBitmap(cvs As AsyncCanvas, obj As TPFDObject) As B4XBitmap
	Dim result As B4XBitmap
	
	'Set the font
	Dim fnt As Font = PFDCommon.SetFontFX(obj)

	'Get the canvas pane 
	'Log($"[TextWrappedToBitmap]1=${cvs.Panel.Width},${cvs.Panel.Height}"$)
	Dim p As Pane = cvs.Panel
	'Log($"[TextWrappedToBitmap]2=${p.Width},${p.Height}"$)
	
	'Create the label used to wrap the text
	Dim lbl As Label
	'Init the label and assign font, textcolor, wraptext flag and the text
	lbl.Initialize("")
	lbl.Font = fnt
	lbl.TextColor = ColorIntToPaint(obj.textcolor)
	lbl.Alignment = TextAlignment2FX(obj.textalignment)
	lbl.Text = obj.value
	lbl.WrapText = True

	'Add the label to the panel - no size required
	p.AddNode(lbl,0,0,0,0)

	'Create a b4xview which is used to create a snapshot as b4xbitmap
	'Assign the previous label with the wrapped text
	Dim xlbl As B4XView = lbl
	'Set the dimensions
	xlbl.Width = obj.width
	xlbl.Height = obj.height
	'xlbl.Color = obj.fillcolor
	xlbl.SetColorAndBorder(obj.fillcolor, obj.strokewidth, obj.bordercolor, obj.cornerradius)
	result = xlbl.Snapshot
	
	'Remove the label from the parent, i.e. the canvas panel
	lbl.RemoveNodeFromParent

	Return result
End Sub

'Convert B4X alignment to FX alignment.
'Only center, left & right are supported.
Public Sub TextAlignment2FX(alignment As String) As String
	Dim result As String
	Select alignment
		Case "LEFT"
			result = "CENTER_LEFT"
		Case "CENTER"
			result = "CENTER"
		Case "RIGHT"
			result = "CENTER_RIGHT"
	End Select
	Return result
End Sub

'Decode a unicode string in format \uXXXX (XXXX is HEX code).
'If no escape sequence \u found, then text is returned as is.
'Examples: Ohm sign U+2126 (\u2126), Micro (mu) U+00B5 (\u00B5).
'Example<code>
'Decode Ohm Sign.
'LabelOhmSign.Text = DecodeUnicodeString("\u2126")
'</code>
Public Sub DecodeUnicodeString(text As String) As String
	Dim sb As StringBuilder
	sb.Initialize
	Dim i As Int = 0

	'Check if text contains unicode escape string \u
	If Not(text.ToLowerCase.Contains("\u")) Then
		Return text
	End If
	
	Try
		Do While i < text.Length
			If text.SubString2(i, i + 2) = "\u" Then
				'Extract the hex code
				Dim hexCode As String = text.SubString2(i + 2, i + 6)
				'Convert hex To decimal
				Dim codePoint As Int = Bit.ParseInt(hexCode, 16)
				'Convert To character
				sb.Append(Chr(codePoint))
				'Move past "\uXXXX"
				i = i + 6
			Else
				'Append regular characters
				sb.Append(text.CharAt(i))
				i = i + 1
			End If
		Loop		
	Catch
		'Catch in case text is entered and string can not be decoded.
		'This is just a placeholder to avoid IDE warning message #9 (not used).
		ToLog(False, $"[ERROR][DecodeUnicodeString] ${LastException}"$)
	End Try
	Return sb.ToString
End Sub
#End Region

#Region Boolean
'Convert state boolean to Yes or No as string.
Public Sub StateToYesNo(state As Boolean) As String
	Return IIf(state, "Yes","No")
End Sub

'Convert state boolean to On or Off as string.
Public Sub StateToOnOff(state As Boolean) As String
	Return IIf(state, "On","Off")
End Sub

'Convert state boolean to 1 or 0 as byte.
Public Sub StateToByte(state As Boolean) As Byte
	Return IIf(state, 1, 0)
End Sub
#End Region

#Region StringBuilder
'Add line of code with prefix TAB (optional) and code + CRLF.
Public Sub AddCodeLine(sb As StringBuilder, settab As Boolean, code As String) As StringBuilder
	If settab Then sb.Append(TAB)
	sb.Append(code).append(CRLF)
	Return sb
End Sub
#End Region

#Region JSON
'Set JSON "key":"value" as string.
Public Sub SetJSONString(key As String, value As String) As String
	Return $""${key}":"${value}""$
End Sub

'Set JSON "key":value as number.
Public Sub SetJSONNumber(key As String, value As String) As String
	Return $""${key}":${value}"$
End Sub
#End Region

#Region Clipboard
' Copy String to the clipboard.
' s - String to copy
Public Sub CopyStringToClipboard(s As String)
	fx.Clipboard.SetString(s)
End Sub

' Paste String from clipboard.
' Return: String
Public Sub PasteStringFromClipboard As String
	Dim result As String = ""
	If fx.Clipboard.HasString Then
		result = fx.Clipboard.GetString
	End If
	Return result
End Sub
#End Region

#Region FontAwesome
'Create an B4XBitmap 32x32px from a fontawsome icon.
'Example<code>
'Icon triangle-exclamation
'Dim bmp as b4xbitmap = FontAwesomeToImage(Chr(0xF071), FontSize, xui.color_magenta)
'</code>
Public Sub FontAwesomeToBitmap(Text As String, FontSize As Float, Color As Int) As B4XBitmap
	Dim xui As XUI
	Dim p As Pane = xui.CreatePanel("")
	p.SetLayoutAnimated(0, 0, 0, 32dip, 32dip)
	Dim cvs1 As B4XCanvas
	cvs1.Initialize(p)
	Dim fnt As B4XFont = xui.CreateFontAwesome(FontSize)
	Dim r As B4XRect = cvs1.MeasureText(Text, fnt)
	Dim BaseLine As Int = cvs1.TargetRect.CenterY - r.Height / 2 - r.Top
	cvs1.DrawText(Text, cvs1.TargetRect.CenterX, BaseLine, fnt, Color, "CENTER")
	Dim b As B4XBitmap = cvs1.CreateBitmap
	cvs1.Release
	Return b
End Sub

'Create an image 32x32px from a fontawsome icon.
Public Sub FontAwesomeToImage (Text As String, FontSize As Float, Color As Int) As Image
	Dim img As Image = FontAwesomeToBitmap (Text, FontSize, Color)
	Return img
End Sub

'Default fontawesome icons for information, warning, error.
'Example<code>
'Dim msg As String = $"This feature is not supported for draw mode select."$
'xui.Msgbox2Async(msg, "[WARNING] Cut Object","OK","","", Helper.FontAwesomeWarning(32))
'</code>
Public Sub FontAwesomeInformation(FontSize As Float) As Image
	Dim text As String = Chr(0xF05A)
	Return FontAwesomeToImage(text, FontSize, xui.color_black)
End Sub
Public Sub FontAwesomeWarning(FontSize As Float) As Image
	Dim text As String = Chr(0xF071)
	Return FontAwesomeToImage(text, FontSize, xui.color_magenta)
End Sub
Public Sub FontAwesomeError(FontSize As Float) As Image
	Dim text As String = Chr(0xF071)
	Return FontAwesomeToImage(text, FontSize, xui.color_red)
End Sub
#End Region

#Region Colors
'Get the shade color from xui color int.
Public Sub ShadeColor(clr As Int) As Int
	Dim argb() As Int = GetARGB(clr)
	Dim factor As Float = 0.75
	Return xui.Color_RGB(argb(1) * factor, argb(2) * factor, argb(3) * factor)
End Sub

'Get the ARGB color from xui color int.
Public Sub GetARGB(Color As Int) As Int()
	Private res(4) As Int
	res(0) = Bit.UnsignedShiftRight(Bit.And(Color, 0xff000000), 24)
	res(1) = Bit.UnsignedShiftRight(Bit.And(Color, 0xff0000), 16)
	res(2) = Bit.UnsignedShiftRight(Bit.And(Color, 0xff00), 8)
	res(3) = Bit.And(Color, 0xff)
	Return res
End Sub

'Convert paint color to xui int color.
Public Sub ColorPaintToInt(clr As Paint) As Int
	Return xui.PaintOrColorToColor(clr)
End Sub

'Convert color hex string ARGB to int.
'The hex string must be 8. If length 6 then FF is set as prefix.
'Example <code>:
''Convert color hex string blue to int value -16776961
'Log(Helper.HexToColor("FF0000FF"))
'</code>
Public Sub ColorHexToInt(hex As String) As Int
	'Log($"[ColorHexToInt]${hex}"$)
	Dim bc As ByteConverter

	If IsNumber(hex) Then
		Dim n As Int = hex
		hex = ColorIntToHexARGB(n)
	End If
	'Remove prefix # or 0x
	If hex.StartsWith("#") Then
		hex = hex.SubString(1)
	Else If hex.StartsWith("0x") Then
		hex = hex.SubString(2)
	End If
	'Check length is 8 else set FF as prefix
	If hex.Length == 6 Then hex = $"FF${hex}"$
	'Convert hex string to bytes
	Dim b() As Byte = bc.HexToBytes(hex)
	'Get the ints from the byte array
	Dim ints() As Int = bc.IntsFromBytes(b)
	'Return the first int array item
	Return ints(0)
End Sub

'Convert xui color value int to hex string AARRGGBB (length 8).
'value - (int) XUI color value.
'Example <code>: xui.color_red converted to FFFF0000.
'Log($"${xui.Color_White} = ${Helper.ColorIntToHexARGB(xui.Color_White)}"$)				'-1
'Log($"${xui.Color_Black} = ${Helper.ColorIntToHexARGB(xui.Color_Black)}"$)				'-16777216
'Log($"${xui.Color_DarkGray} = ${Helper.ColorIntToHexARGB(xui.Color_DarkGray)}"$)		'-12303292
'Log($"${xui.Color_Lightgray} = ${Helper.ColorIntToHexARGB(xui.Color_LightGray)}"$)		'-3355444
'</code>
Public Sub ColorIntToHexARGB(value As Int) As String
	Dim bc As ByteConverter
	Return bc.HexFromBytes(bc.IntsToBytes(Array As Int(value)))
End Sub

'Convert xui color value int to hex string RRGGBB (length 6)
'value - (int) XUI color value.
'Example: xui.color_red converted to FF0000.
Public Sub ColorIntToHexRGB(value As Int) As String
	Dim bc As ByteConverter
	Return bc.HexFromBytes(bc.IntsToBytes(Array As Int(value))).SubString(2)
End Sub

'Convert xui color value int to paint.
'value - (int) XUI color value.
'Example: xui.color_rgb(221,221,221) converted to 0xddddddff.
Public Sub ColorIntToPaint(value As Int) As Paint
	Return fx.Colors.From32Bit(value)
End Sub
#End Region

#Region Images
'Get the height and width in px of an PNG image.
'Returns map with keys h and w or an uninitialized map if error.
'Example<code>
'Dim m As Map = Helper.GetImageSize(File.dirapp, TextFieldObjImage.Text)
'If m.IsInitialized Then
'	Log($"Image Size ${File.dirapp}\${TextFieldObjImage.Text}:${CRLF}width=${m.Get("w")}${CRLF}height=${m.Get("h")}"$)
'Else
'	Log($""[ERROR] Can not read image ${File.dirapp}\${TextFieldObjImage.Text}"$)
'End If
'</code>
Public Sub GetImageSize(folder As String, filename As String) As Map
	Dim m As Map
	Dim h As Int
	Dim w As Int
	
	'Check image file exists. If not found return uninitialised map.
	If Not(File.Exists(folder, filename)) Then
		Return m
	End If
	
	'Define the File JavaObject
	Dim joFile As JavaObject
	'Init file object with new instance from file path
	joFile.InitializeNewInstance("java.io.File", Array As Object($"${folder}\${filename}"$))
	
	'Convert the byte array input stream into buffered image from the file javaobject
	Dim joImageIO As JavaObject
	joImageIO.InitializeStatic("javax.imageio.ImageIO")
	Dim joBufferedImage As JavaObject = joImageIO.RunMethod("read", Array(joFile))
	h = joBufferedImage.RunMethod("getHeight", Null)
	w = joBufferedImage.RunMethod("getWidth", Null)

	'Create & return the map with the two keys
	m = CreateMap("h":h, "w":w)		
	Return m
End Sub
#End Region

#Region JSON
'Convert a map to JSOn string.
'{"description":"Decision","tag":"DECISION-CONNECT-BT"}
Public Sub MapToJSON(m As Map) As String
	Dim result As String = ""
	Dim json As JSONGenerator
	
	'Init the JSon string and create the result string
	Try
		json.Initialize(m)
		result = json.ToString			
	Catch
		Log($"[ERROR][MapToJSON]${LastException}"$)
	End Try
	Return result
End Sub

'Get JSON key value as string from JSON string.
'Arrays are not supported.
Public Sub GetJSONKeyValue(key As String, json As String) As String
	Dim result As String = ""
	Dim parser As JSONParser
		
	'Init the JSon string and create the result string
	Try
		parser.Initialize(json)
		Dim obj As Map = parser.NextObject
		result = obj.Get(key)
	Catch
		Log($"[ERROR][GetJSONKeyValue]${LastException}"$)
	End Try
	Return result
End Sub

#End Region
