﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=10
@EndOfDesignText@
#Region Class Information
'Class: PFDBase
'File: PFDBase.bas
'Base class to define & draw objects the canvas.
'
'PFDCreator Copyright see B4XMainPage.
#End Region

Sub Class_Globals
	Private debug As Boolean
	
	'Core
	Private xui As XUI
	Private fx As JFX

	'Canvas
	Public mCvs As AsyncCanvas	'AsyncCanvas

	'PFD Object Type
	'Defines the properties of an object.
	'The properties are set depending the object type. An object might not use all properties.
	Type TPFDObject (index As Int, _					'Unique index of the object - value is set dynamic
					 pfdobject As String, _ 			'Type of the object - set in pfdcreator.shp
					 tag As String, _ 					'Unique mandatory tag
					 description As String, _ 			'Brief description of the object
					 value As String, _ 
					 unit As String, _ 
					 direction As String, _ 
					 arrow As String, _ 
					 centerx As Float, _ 
					 centery As Float, _ 
					 width As Float, _
					 height As Float, _ 
					 x1 As Float, _ 
					 y1 As Float, _
					 x2 As Float, _
					 y2 As Float, _ 
					 image As String, _ 
					 font As String, _
					 fontsize As Int, _ 
					 fillcolor As Int, _
					 bordercolor As Int, _ 
					 textcolor As Int, _
					 textalignment As String, _ 
					 strokewidth As Int, _
					 cornerradius As Int, _ 
					 separator As Boolean, _ 
					 operate As Boolean, _ 
					 rect(5) As TPFDRect, _ 
					 shapeselector As String)

	'Type for a point X,Y position on the canvas.
	'Optional a point id can be assigned.
	Type TPFDPoint (id As Int, x As Int, y As Int)

	'Type for the PFDObject rectangle: x1,y1=topleft, x2,y2=botright.
	'A PFDObject has 5 rectangles:
	'id=0 for the main rect covering the shape
	'id=1-4 for the resize rect topleft (1), botleft(2), topright(3), botright(4)
	Type TPFDRect (id As Int, x1 As Float, y1 As Float, x2 As Float, y2 As Float)

	'Type for a line	
	Type TPFDLine (x1 As Float,y1 As Float,x2 As Float,y2 As Float,fillcolor As Int, strokewidth As Float)
	
	'Type for an arrow
	Type TPFDArrow (cx As Float, cy As Float, w As Float, h As Float, direction As String, fillcolor As Int)
End Sub

'Initializes the object.
'cvs - Canvas used as PFD Drawing Board.
Public Sub Initialize(cvs As AsyncCanvas, debugmode As Boolean)
	mCvs = cvs
	debug = debugmode
End Sub

#Region ReDraw
'Draw the pfd from the external pfd definition file json format.
Public Sub ReDraw(pfdobjects As List, grid As Boolean, color As Int)
	'Clear Canvas
	mCvs.Clear(color)

	'Draw the grid
	If grid Then PFDCommon.DrawGrid(mCvs, False)

	'Loop over the pfdobjects, select & draw
	
	'Draw the basic shapes first!
	PFDShapesBasic.DrawShapesFromList(mCvs, pfdobjects)
	
	'Draw the PFD Specific Types
	PFDShapesPID.DrawShapesFromList(mCvs, pfdobjects)
	PFDShapesFlowchart.DrawShapesFromList(mCvs, pfdobjects)
	PFDShapesElectrical.DrawShapesFromList(mCvs, pfdobjects)
	'Add more
	
	'Commit the drawings
	mCvs.Invalidate
End Sub
#End Region

#Region PFDObjects
'Create the PFD objects list from PFD definition JSON array key "objects".
'objects - list of objects parsed from the pfd file.
Public Sub GetObjectsFromPFDDefinition(objects As List) As List
	Dim result As List

	result.Initialize
	
	'Parse the pfd
	Try
		For Each coljRoot As Map In objects
			'Define and init a new PFDObject
			Dim pfdobject As TPFDObject
			pfdobject.Initialize
			
			pfdobject.index 		= coljRoot.Get("index")
			pfdobject.pfdobject 	= coljRoot.Get("pfdobject")
			pfdobject.tag 			= coljRoot.Get("tag")
			If pfdobject.tag.Trim.Length == 0 Then
				Log($"[WARNING][GetObjectsFromPFDDefinition] Object has no mandatory tag. ${coljRoot}"$)
			End If
			pfdobject.description	= coljRoot.Get("description")
			pfdobject.value			= coljRoot.Get("value")
			pfdobject.unit			= coljRoot.Get("unit")
			pfdobject.direction		= coljRoot.Get("direction")
			pfdobject.arrow 		= coljRoot.Get("arrow")

			pfdobject.centerx 		= coljRoot.Get("centerx")
			pfdobject.centery 		= coljRoot.Get("centery")
			pfdobject.width 		= coljRoot.Get("width")
			pfdobject.height 		= coljRoot.Get("height")
			'Obj rectangle dimension topleft (x1,y2) to botleft (x2,y2) or begin to end for a line
			pfdobject.x1 			= IIf(coljRoot.ContainsKey("x1"), coljRoot.Get("x1"), 0)
			pfdobject.y1 			= IIf(coljRoot.ContainsKey("y1"), coljRoot.Get("y1"), 0)
			pfdobject.x2 			= IIf(coljRoot.ContainsKey("x2"), coljRoot.Get("x2"), 0)
			pfdobject.y2 			= IIf(coljRoot.ContainsKey("y2"), coljRoot.Get("y2"), 0)

			pfdobject.image			= IIf(coljRoot.ContainsKey("image"), coljRoot.Get("image"), "")
			pfdobject.font			= IIf(coljRoot.ContainsKey("font"), coljRoot.Get("font").As(String).ToUpperCase, PFDCommon.FONT_DEFAULT)
			pfdobject.fontsize		= IIf(coljRoot.ContainsKey("fontsize"), coljRoot.Get("fontsize"), PFDCommon.FONT_DEFAULT_SIZE)
			pfdobject.fillcolor		= IIf(coljRoot.ContainsKey("fillcolor"), Helper.ColorHexToInt(coljRoot.Get("fillcolor")), xui.color_white)
			pfdobject.bordercolor	= IIf(coljRoot.ContainsKey("bordercolor"), Helper.ColorHexToInt(coljRoot.Get("bordercolor")), xui.color_black)
			pfdobject.textcolor		= IIf(coljRoot.ContainsKey("textcolor"), Helper.ColorHexToInt(coljRoot.Get("textcolor")), xui.color_black)
			pfdobject.textalignment	= IIf(coljRoot.ContainsKey("textalignment"), coljRoot.Get("textalignment"), PFDCommon.TEXT_ALIGNMENT_CENTER)
			pfdobject.strokewidth	= IIf(coljRoot.ContainsKey("strokewidth"), coljRoot.Get("strokewidth"), 2)
			pfdobject.cornerradius	= IIf(coljRoot.ContainsKey("cornerradius"), coljRoot.Get("cornerradius"), 0)
			pfdobject.separator		= IIf(coljRoot.ContainsKey("separator"), coljRoot.Get("separator"), False)
			pfdobject.operate		= IIf(coljRoot.ContainsKey("operate"), coljRoot.Get("operate"), False)

			'Obj selected 5 rectangles center + 4 corners
			pfdobject = SetPFDObjectRectangles(pfdobject)
			
			Helper.ToLog(debug, $"obj=${pfdobject.pfdobject},tag=${pfdobject.tag},cx=${pfdobject.centerx},cy=${pfdobject.centery}"$)
			
			'Add the PFD object to the pfdobjects list
			result.add(pfdobject)
		Next
	
	Catch
		PFDCommon.ShowMessage(PFDCommon.MSG_TYPE_ERROR, _
		                 	  "PFD Definition Get Objects", _ 
							  $"Can not parse the JSON array of objects.${CRLF}${LastException}"$)
	End Try

	'Return the pfd objects list
	Return result
End Sub

'Set the 5 PFDObject rectangles using the PFDObject centerx,y pos and the width and height.
Public Sub SetPFDObjectRectangles(pfdobject As TPFDObject) As TPFDObject
	'PFDObject rectangles
	'Main Rect id=0
	pfdobject.rect(0).Initialize
	pfdobject.rect(0).id = 0
	pfdobject.rect(0).x1 = pfdobject.centerx - (pfdobject.width/2)
	pfdobject.rect(0).y1 = pfdobject.centery - (pfdobject.height/2)
	pfdobject.rect(0).x2 = pfdobject.centerx + (pfdobject.width/2)
	pfdobject.rect(0).y2 = pfdobject.centery + (pfdobject.height/2)

	'Topleft id=1
	pfdobject.rect(1).Initialize
	pfdobject.rect(1).id = 1
	pfdobject.rect(1).x1 = pfdobject.centerx - (pfdobject.width/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(1).y1 = pfdobject.centery - (pfdobject.height/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(1).x2 = pfdobject.rect(1).x1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)
	pfdobject.rect(1).y2 = pfdobject.rect(1).y1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)

	'Botleft id=2
	pfdobject.rect(2).Initialize
	pfdobject.rect(2).id = 2
	pfdobject.rect(2).x1 = pfdobject.centerx - (pfdobject.width/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(2).y1 = pfdobject.centery + (pfdobject.height/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(2).x2 = pfdobject.rect(2).x1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)
	pfdobject.rect(2).y2 = pfdobject.rect(2).y1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)

	'Topright id=3
	pfdobject.rect(3).Initialize
	pfdobject.rect(3).id = 3
	pfdobject.rect(3).x1 = pfdobject.centerx + (pfdobject.width/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(3).y1 = pfdobject.centery - (pfdobject.height/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(3).x2 = pfdobject.rect(3).x1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)
	pfdobject.rect(3).y2 = pfdobject.rect(3).y1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)

	'Botright id=4
	pfdobject.rect(4).Initialize
	pfdobject.rect(4).id = 4
	pfdobject.rect(4).x1 = pfdobject.centerx + (pfdobject.width/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(4).y1 = pfdobject.centery + (pfdobject.height/2) - PFDCommon.PFDOBJECT_CORNERRECTSIZE
	pfdobject.rect(4).x2 = pfdobject.rect(4).x1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)
	pfdobject.rect(4).y2 = pfdobject.rect(4).y1 + (PFDCommon.PFDOBJECT_CORNERRECTSIZE*2)

	Return pfdobject
End Sub

'Check if a rect on the canvas is touched.
Public Sub IsRectTouched(x As Float, y As Float, rect As TPFDRect) As Boolean
	Dim x1,y1,x2,y2 As Double
	Dim offset As Int = 2
	Dim result As Boolean = False
			
	'Check if object touched in object rectangle area
	x1 = rect.x1 - offset
	y1 = rect.y1 - offset
	x2 = rect.x2 + offset
	y2 = rect.y2 + offset
	If x >= x1 And x <= x2 And _
	   y >= y1 And y <= y2 Then
		result = True
	End If
	Return result
End Sub

'Get the object touched from the x,y touch position.
'objects - List of objects from which the object is selected.
'x,y - X,Y coordinates of the PFD Drawing Board (canvas) touch position
'An offset, based on the object strokewidth, is used to increase the rectangle touch range to better touch objects esp. lines.
Public Sub GetObjectTouched(objects As List, x As Float, y As Float) As TPFDObject
	For Each obj As TPFDObject In objects
		Helper.ToLog(debug, $"[GetObjectTouched]x=${x}, y=${y}, tag=${obj.tag}, 0: id=${obj.rect(0).id}, x1=${obj.rect(0).x1}, y1=${obj.rect(0).y1}, x2=${obj.rect(0).x2}, y2=${obj.rect(0).y2}"$)
		'LogRetTouched(obj, x, y)
		For Each r As TPFDRect In obj.rect
			If IsRectTouched(x, y, r) Then
				Helper.ToLog(debug, $"[GetObjectTouched TRUE]index=${obj.index},tag=${obj.tag},rect: id=${r.id},x=${x},y=${y},x1=${r.x1},y1=${r.y1},x2=${r.x2},y2=${r.y2}"$)
				Return obj
			End If
		Next
	Next
	Return Null
End Sub

'Get the object rect touched from the x,y touch position.
'obj - Object touched.
'x,y - X,Y coordinates of the PFD Drawing Board (canvas) touch position
'Returns 0 for the main rect and 1-4 for the corner rect.
Public Sub GetObjectRectTouched(obj As TPFDObject, x As Float, y As Float) As Int
	Dim result As Int = -1
	
	'Loop over the object rect to find the rect touched - start with higest rect
	Dim i As Int = 4
	Do While i > -1
		If IsRectTouched(x, y, obj.rect(i)) Then
			Helper.ToLog(debug, $"[GetObjectTouched TRUE]index=${obj.index},tag=${obj.tag},rect: id=${obj.rect(i).id},x=${x},y=${y},x1=${obj.rect(i).x1},y1=${obj.rect(i).y1},x2=${obj.rect(i).x2},y2=${obj.rect(i).y2}"$)
			Return obj.rect(i).id
		End If
		i = i - 1
	Loop
	Return result
End Sub

'Resize an object for the selected rect id (see PFDCommon DrawObjectSelected).
'Example:
'Moving the bottom right corner to the left down resized the obj and sets new centerx,y & width+height.
'The left top&bottom corners remain at its pos, but topleft is moved right and botright is moved right and down.
'obj - Object selected.
'rectid - Selected rect from the 5 rects (center + 4 corners) around the object.
'x,y - Mouse position.
Public Sub ResizeObjectTouched(obj As TPFDObject, rectid As Int, x As Float, y As Float) As TPFDObject
	'Helper for the obj rect size
	Dim x1,y1,x2,y2 As Float
	Dim cx,cy,dx,dy As Float
	
	'New x and y pos
	x = Round(X / PFDCommon.GRID_SIZE) * PFDCommon.GRID_SIZE
	y = Round(Y / PFDCommon.GRID_SIZE) * PFDCommon.GRID_SIZE

	'Define the cur obj rect size
	cx = obj.centerx
	cy = obj.centery
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)

	'Select one of the 5 rect and set obj new centerx,y & width+height
	Select rectid
		Case PFDCommon.PFDOBJECT_RECT_CENTER
			'No action because center rect can not be resized
		Case PFDCommon.PFDOBJECT_RECT_TOPLEFT
			'Get the delta in size
			dx = x1 - x
			dy = y1 - y
			'Set the new w,h
			obj.width = x2 - x
			obj.height = y2 - y
			'Set the new centerx,y
			obj.centerx = cx - (dx/2)
			obj.centery = cy - (dy/2)
		Case PFDCommon.PFDOBJECT_RECT_BOTLEFT
			'Get the delta in size
			dx = x1 - x
			dy = y - y2			
			'Set the new w,h
			obj.width = x2 - x
			obj.height = y - y1
			'Set the new centerx,y
			obj.centerx = cx - (dx/2)
			obj.centery = cy + (dy/2)
		Case PFDCommon.PFDOBJECT_RECT_TOPRIGHT
			'Get the delta in size
			dx = x - x2
			dy = y1 - y
			'Set the new w,h
			obj.width = x - x1
			obj.height = y2 - y
			'Set the new centerx,y
			obj.centerx = cx + (dx/2)
			obj.centery = cy - (dy/2)
		Case PFDCommon.PFDOBJECT_RECT_BOTRIGHT
			'Get the delta in size
			dx = x - x2
			dy = y - y2
			'Set the new w,h
			obj.width = x - x1
			obj.height = y - y1
			'Set the new centerx,y
			obj.centerx = cx + (dx/2)
			obj.centery = cy + (dy/2)			
	End Select
	'Return the updated obj
	Return obj
End Sub

'Get an operate object touched from the x,y touch position.
'objects - List of objects from which the object is selected.
'x,y - X,Y coordinates of the PFD Drawing Board (canvas) touch position
'An offset, based on the object strokewidth, is used to increase the rectangle touch range to better touch objects esp. lines.
Public Sub GetOperateObjectTouched(objects As List, x As Float, y As Float) As TPFDObject
	Dim x1,y1,x2,y2 As Double
	Dim offset As Int
	
	For Each obj As TPFDObject In objects
		'Check if object touched in rectangle area
		offset = obj.strokewidth * 2
		x1 = obj.centerx - (obj.width / 2) - offset
		y1 = obj.centery - (obj.height / 2) - offset
		x2 = x1 + obj.width + offset
		y2 = y1 + obj.height + offset
		If x >= x1 And x <= x2 And _
		   y >= y1 And y <= y2 And obj.operate Then
		   	Helper.ToLog(debug, $"[GetObjectOperateTouched TRUE]index=${obj.index},${obj.tag},x=${x},y=${y},x1=${x1},y1=${y1},x2=${x2},y2=${y2}"$)
			Return obj
		End If
	Next
	Return Null
End Sub

'Check if an object is not null and is initialzed.
'Returns True if OK
Public Sub IsObjectInitialized(obj As TPFDObject) As Boolean
	Dim result As Boolean = True
	'Check if an object is selected = there must be initialized draggingobject
	If obj == Null Then 
		Return False
	End If
	If Not(obj.IsInitialized) Then 
		Return False
	End If
	Return result
End Sub

'Update an object in the objects list by the object index.
'Note: The index of an object is set to a new value by PFDCreator.
'It is recommended to use the unique tag.
'objects - List of PFD objects
'obj - Object to be updated in the list using the object index
'Returns updated objects list
Public Sub UpdateObjectByIndex(objects As List, obj As TPFDObject) As List
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object index matches the given index
		If objects.Get(i).As(TPFDObject).index == obj.index Then
			'Replace the object in the list with the obj given
			objects.Set(i, obj)
		End If
	Next
	'Return updated list
	Return objects
End Sub

'Update an object in the objects list by its unique object tag.
'objects - List of PFD objects
'tag - Object unique tag
'value - Object new value
'Returns updated objects list
'Example: <code>
'Set the value of the object with tag I-1 to 1234
'PFDObjects = PFD.UpdateObjectValueByTag(PFDObjects, "I-1", 1234)
'</code>
Public Sub UpdateObjectByTag(objects As List, obj As TPFDObject) As List
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object tag matches the given tag
		If objects.Get(i).As(TPFDObject).tag.ToUpperCase == obj.tag.ToUpperCase Then
			'Replace the object in the list
			objects.Set(i, obj)
			'Log($"[UpdateObjectValueByTag] replaced i=${i}, tag=${obj.tag}, value=${obj.value}"$)
		End If
	Next
	'Return updated list
	Return objects
End Sub

'Update an object value in the objects list by its unique object tag.
'objects - List of PFD objects
'tag - Object unique tag
'value - Object new value
'Returns updated objects list
'Example: <code>
'Set the value of the object with tag I-1 to 1234
'PFDObjects = PFD.UpdateObjectValueByTag(PFDObjects, "I-1", 1234)
'</code>
Public Sub UpdateObjectValueByTag(objects As List, tag As String, value As String) As List
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object tag matches the given tag
		If objects.Get(i).As(TPFDObject).tag.ToUpperCase == tag.ToUpperCase Then
			'Get the object
			Dim obj As TPFDObject = objects.Get(i).As(TPFDObject)
			'Set the new value
			obj.value = value
			'Replace the object in the list
			objects.Set(i, obj)
			'Log($"[UpdateObjectValueByTag] replaced i=${i}, tag=${obj.tag}, value=${obj.value}"$)
		End If
	Next
	'Return updated list
	Return objects
End Sub

'Get the next free object index.
Public Sub GetNextIndex(objects As List) As Int
	Dim result As Int = -1
	'Loop over the objects list
	For Each obj As TPFDObject In objects
		If obj.index > result Then result = obj.index		
	Next
	Return result + 1
End Sub

'Copy an existing object to a new object with new index and tag
Public Sub CopyObject(obj As TPFDObject) As TPFDObject
	Dim newobj As TPFDObject
	newobj.Initialize
	newobj.arrow = obj.arrow
	newobj.bordercolor = obj.bordercolor
	newobj.centerx = 100
	newobj.centery = 100
	newobj.cornerradius = obj.cornerradius
	newobj.description = obj.description
	newobj.direction = obj.direction
	newobj.fillcolor = obj.fillcolor
	newobj.font = obj.font
	newobj.fontsize = obj.fontsize
	newobj.height = obj.height
	newobj.image = obj.image
	newobj.index = -1	'GetNextIndex(objects)
	newobj.operate = obj.operate
	newobj.pfdobject = obj.pfdobject
	newobj.separator = obj.separator
	newobj.strokewidth = obj.strokewidth
	newobj.tag = $"${obj.tag}-2"$
	newobj.textalignment = obj.textalignment
	newobj.textcolor = obj.textcolor
	newobj.unit = obj.unit
	newobj.value = obj.value
	newobj.width = obj.width
	newobj.x1 = obj.x1
	newobj.y1 = obj.y1
	newobj.x2 = obj.x2
	newobj.y2 = obj.y2
	newobj = SetPFDObjectRectangles(newobj)
	Return newobj
End Sub

'Get the list of PFD Objects from the PFD Drawing Board selection rectangle.
'The method GetObjectTouched is used to select the object(s).
'objects - list of selected objects
Public Sub GetObjectsFromSelectArea(objects As List, rect As B4XRect) As List
	Dim result As List

	'Init the results list
	result.Initialize
	
	'Go over the rect dimension
	Dim x,y As Int
	For x = rect.left To rect.right Step PFDCommon.GRID_SIZE
		For y = rect.top To rect.bottom Step PFDCommon.GRID_SIZE
			Dim obj As TPFDObject = GetObjectTouched(objects, x, y)
			If obj <> Null Then
				If result.IndexOf(obj) == -1 Then result.Add(obj)
			End If
		Next
	Next

	'Return the pfd objects list
	Return result
End Sub

'Get an object from the objects list by the object index.
Public Sub GetObjectByIndex(objects As List, obj As TPFDObject) As TPFDObject
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object index matches the given index
		If objects.Get(i).As(TPFDObject).index == obj.index Then
			Return obj
		End If
	Next
	Return Null
End Sub

'Get an object from the objects list by the tag.
Public Sub GetObjectByTag(objects As List, tag As String) As TPFDObject
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object tag matches the given tag
		Dim obj As TPFDObject = objects.Get(i).As(TPFDObject)
		If obj.tag == tag Then
			Return obj
		End If
	Next
	Return Null
End Sub

'Get the value of an object from the objects list by the tag.
Public Sub GetObjectValueByTag(objects As List, tag As String) As String
	'Loop over the objects list
	For i = 0 To objects.Size - 1
		'Check if object tag matches the given tag
		Dim obj As TPFDObject = objects.Get(i).As(TPFDObject)
		If obj.tag == tag Then
			Return obj.value
		End If
	Next
	Return Null
End Sub

'USED FOR PFDOperator
'Get the list of operate objects from the PFD Objects list.
'Only the object tag is added.
Public Sub GetOperateObjects(objects As List) As List
	Dim result As List
	result.Initialize
	For Each obj As TPFDObject In objects
		If obj.operate Then
			result.Add(obj.tag)
		End If
	Next
	Return result
End Sub
#End Region

#Region PFD Shapes
'Check if a shape JSON key exists in the shape definition file pfdcreator.shp.
'The global list PFDShapes contains the PFD shapes.
'obj - Shape type name (pfdobject)
'key - Shape property key
'Example<code>
'Dim haskey as boolean = HasShapeProperty(obj, "cornerradius")
'</code>
Public Sub HasShapeProperty(shapes As List, obj As String, key As String) As Boolean
	Dim result As Boolean = False
	
	'Loop over the shapes
	For Each colshape As Map In shapes
		'Shape found for the obj
		If colshape.Get("pfdobject") == obj Then
			Return IIf(colshape.ContainsKey(key), True, False)
		End If
	Next
	Return result
End Sub

'Get the hint (if exists) for the shape selected.
'The shape hint is defined in the shape definition file pfdcreator.shp as JSON key hint.
'The global list PFDShapes contains the PFD shapes.
'obj - Shape type name (pfdobject)
Public Sub GetShapeHint(shapes As List, obj As String) As String
	Dim result As String = ""
	
	For Each colshape As Map In shapes
		If colshape.Get("pfdobject") == obj Then
			Return IIf(colshape.ContainsKey("hint"), colshape.Get("hint"), "")
		End If
	Next
	Return result
End Sub

'Get the method (if exists) for the shape selected.
'The shape method is defined in the shape definition file pfdcreator.shp as JSON key method.
'The global list PFDShapes contains the PFD shapes.
'obj - Shape type name (pfdobject)
Public Sub GetShapeMethod(shapes As List, obj As String) As String
	Dim result As String = ""
	
	For Each colshape As Map In shapes
		If colshape.Get("pfdobject") == obj Then
			Return IIf(colshape.ContainsKey("method"), colshape.Get("method"), "")
		End If
	Next
	Return result
End Sub
#End Region

#Region PFD Definition File
'Get the PFD file type as string from the PFD definition file.
'pfdfile (string) - PFD filename with extension .pfd
'Returns - (string) pfd type like flowchart, pid
Public Sub GetPFDTypeFromFile(pfdfile As String) As String
	Dim result As String
	
	If File.Exists(File.DirApp, pfdfile) Then
		Dim pfdfilecontent As String = File.ReadString(File.DirApp, pfdfile)
		Dim parser As JSONParser
		parser.Initialize(pfdfilecontent)

		Dim jRoot As Map = parser.NextObject
		result = jRoot.Get("pfdtype")
	End If
	Return result
End Sub

'Get the PFD description as string from the PFD definition file.
'pfdfile (string) - PFD filename with extension .pfd
'Returns - (string) pfd description 
Public Sub GetPFDDescriptionFromFile(pfdfile As String) As String
	Dim result As String
	
	If File.Exists(File.DirApp, pfdfile) Then
		Dim pfdfilecontent As String = File.ReadString(File.DirApp, pfdfile)
		Dim parser As JSONParser
		parser.Initialize(pfdfilecontent)

		Dim jRoot As Map = parser.NextObject
		result = IIf(jRoot.ContainsKey("pfddescription"), jRoot.Get("pfddescription"), "")
	End If
	Return result
End Sub

'Get the PFD lastchange as string from the PFD definition file.
'pfdfile (string) - PFD filename with extension .pfd
'Returns - (string) pfd last change
Public Sub GetPFDLastChangeFromFile(pfdfile As String) As String
	Dim result As String
	
	If File.Exists(File.DirApp, pfdfile) Then
		Dim pfdfilecontent As String = File.ReadString(File.DirApp, pfdfile)
		Dim parser As JSONParser
		parser.Initialize(pfdfilecontent)

		Dim jRoot As Map = parser.NextObject
		result = IIf(jRoot.ContainsKey("pfdlastchange"), jRoot.Get("pfdlastchange"), "")
	End If
	Return result
End Sub

'Get the PFD background color as int from the PFD definition file.
'If the key pfdbackgroundcolor not exists, a default is returned (PFDCommon.COLOR_ISA101_SILVER192)
'pfdfile (string) - PFD filename with extension .pfd
'Returns - (int) pfd background color
Public Sub GetPFDBackgroundColorFromFile(pfdfile As String) As Int
	Dim result As Int
	
	If File.Exists(File.DirApp, pfdfile) Then
		Dim pfdfilecontent As String = File.ReadString(File.DirApp, pfdfile)
		Dim parser As JSONParser
		parser.Initialize(pfdfilecontent)

		Dim jRoot As Map = parser.NextObject
		'Check if the pfdbackgroundcolor key exists
		result = IIf(jRoot.ContainsKey("pfdbackgroundcolor"), _ 
				     Helper.ColorHexToInt(jRoot.Get("pfdbackgroundcolor")), _ 
					 Helper.ColorHexToInt(PFDCommon.COLOR_ISA101_WHITE))
	End If
	Return result
End Sub

'Get the PFD file objects into a list from the PFD definition file.
'pfdfile (string) - PFD filename with extension .pfd
'Returns - (string) JSON array object holding the PFD definition
Public Sub GetObjectsFromFile(pfdfile As String) As List
	Dim result As List
	
	If File.Exists(File.DirApp, pfdfile) Then
		Dim pfdfilecontent As String = File.ReadString(File.DirApp, pfdfile)
		Dim parser As JSONParser
		parser.Initialize(pfdfilecontent)

		Dim jRoot As Map = parser.NextObject
		'Dim PFDType As String = jRoot.Get("pfdtype")
		result = jRoot.Get("objects")
	End If
	Return result
End Sub

'Save the list of objects to a PFD file in the dirapp (files) folder.
'The PFD file has JSON format.
'{"pfddescription":"sws test flow","objects":[
'	{"cornerradius":0,"arrow":"RIGHT","textcolor":-16777216,"fontsize":12,"y1":0,"y2":0,"textalignment":"CENTER","tag":"LSH-1","value":"SolidLine","direction":"HORIZONTAL","height":1,"image":"","centery":179,"centerx":500,"index":0,"separator":False,"bordercolor":-16777216,"unit":"","fillcolor":-16777216,"operate":False,"pfdobject":"LINESOLIDHORIZONTAL","width":360,"x1":0,"x2":0,"desc":"Solid line","font":"DEFAULT","strokewidth":2},
'	...
'	{"cornerradius":0,"arrow":"NONE","textcolor":-1,"fontsize":20,"y1":0,"y2":0,"textalignment":"CENTER","tag":"TB-4","value":"Textbox","direction":"NONE","height":80,"image":"","centery":350,"centerx":750,"index":7,"separator":False,"bordercolor":-16777216,"unit":"","fillcolor":-65536,"operate":False,"pfdobject":"TEXTBOX","width":140,"x1":0,"x2":0,"desc":"Textbox","font":"DEFAULTBOLD","strokewidth":2}
']}
'pfdtype - pid, flowchart
'pfddescription - Short description
'object - list of objects with TPFDObject items
'pfdfolder - dirapp folder (=File.DirApp)
'pfdfile - pfd filename with extension .pfd (=PFD_DEFINITION_FILE)
'pfdbackgroundcolor - pfd background color which is given as int but stored as hex ARGB
Public Sub SaveToFile(pfdfile As String, objects As List, pfddescription As String, pfdlastchange As String, pfdbackgroundcolor As Int) As Boolean
	Dim result As Boolean
	Dim sb As StringBuilder

	'Init the stringbuilder
	sb.Initialize
	sb.Append("{").Append(CRLF)
	'Add the globals
	sb.Append(Helper.SetJSONString("pfddescription", pfddescription)).Append(",").Append(CRLF)
	sb.Append(Helper.SetJSONString("pfdlastchange", pfdlastchange)).Append(",").Append(CRLF)
	sb.Append(Helper.SetJSONString("pfdbackgroundcolor", Helper.ColorIntToHexARGB(pfdbackgroundcolor))).Append(",").Append(CRLF)
	
	'Add the objects which is an array with maps holding the type tpfdobjects
	sb.Append($""objects""$).append(":").Append(CRLF)

	sb.Append("[").Append(CRLF)
	'Loop over the object and add a map with pfd object properties
	'Sort the object list by index 0 - N
	objects.SortType("index",True)
	For Each o As TPFDObject In objects
		sb.Append("{")
		sb.Append(Helper.SetJSONNumber("index", o.index)).Append(",")
		sb.Append(Helper.SetJSONString("pfdobject", o.pfdobject)).Append(",")
		sb.Append(Helper.SetJSONString("tag", o.tag)).Append(",")
		sb.Append(Helper.SetJSONString("description", o.description)).Append(",")
		sb.Append(Helper.SetJSONString("value", o.value)).Append(",")
		sb.Append(Helper.SetJSONString("unit", o.unit)).Append(",")
		sb.Append(Helper.SetJSONString("direction", o.direction)).Append(",")
		sb.Append(Helper.SetJSONString("arrow", o.arrow)).Append(",")
		sb.Append(Helper.SetJSONString("image", o.image)).Append(",")
		sb.Append(Helper.SetJSONNumber("centerx", o.centerx)).Append(",")
		sb.Append(Helper.SetJSONNumber("centery", o.centery)).Append(",")
		sb.Append(Helper.SetJSONNumber("width", o.width)).Append(",")
		sb.Append(Helper.SetJSONNumber("height", o.height)).Append(",")
		sb.Append(Helper.SetJSONNumber("x1", o.x1)).Append(",")
		sb.Append(Helper.SetJSONNumber("y1", o.y1)).Append(",")
		sb.Append(Helper.SetJSONNumber("x2", o.x2)).Append(",")
		sb.Append(Helper.SetJSONNumber("y2", o.y2)).Append(",")
		sb.Append(Helper.SetJSONString("font", o.font)).Append(",")
		sb.Append(Helper.SetJSONNumber("fontsize", o.fontsize)).Append(",")
		sb.Append(Helper.SetJSONString("fillcolor", Helper.ColorIntToHexARGB(o.fillcolor))).Append(",")
		sb.Append(Helper.SetJSONString("bordercolor", Helper.ColorIntToHexARGB(o.bordercolor))).Append(",")
		sb.Append(Helper.SetJSONString("textcolor", Helper.ColorIntToHexARGB(o.textcolor))).Append(",")
		sb.Append(Helper.SetJSONString("textalignment", o.textalignment)).Append(",")
		sb.Append(Helper.SetJSONNumber("strokewidth", o.strokewidth)).Append(",")
		sb.Append(Helper.SetJSONNumber("cornerradius", o.cornerradius)).Append(",")
		sb.Append(Helper.SetJSONString("separator", o.separator)).Append(",")
		sb.Append(Helper.SetJSONString("operate", o.operate))
		sb.Append("},").Append(CRLF)
	Next

	'Add the array close ] and JSON close character }.
	Dim content As String = $"${sb.ToString}]${CRLF}}"$
	'Remove the last comma (,) to ensure proper JSON structure: ,]} to ]}
	content = content.Replace($",${CRLF}]"$, $"${CRLF}]"$)

	'Write the string content to the file
	Try
		File.WriteString(File.DirApp, pfdfile, content)
		'xui.Msgbox2Async($"PFD file ${pfdfile} saved"$, _
		'                 "Save PFD", _ 
		'				 "OK","","", Helper.FontAwesomeInformation(24))
		result = True
	Catch
		PFDCommon.ShowMessage(PFDCommon.MSG_TYPE_ERROR, _
		                 	  "PFD Save", _ 
							  $"Can not save the PFD to file ${pfdfile}.${CRLF}${LastException}"$)
		result = False
	End Try
	Return result
End Sub
#End Region

#Region Export
'Export the PFD canvas to a PNG impage.
'pfdfile - PFD filename with extension, i.e. myplant.pfd (is converted to lowercase)
Public Sub ExportPFDToPNG(pfdfile As String)
	Dim pfdbitmap As B4XBitmap = mCvs.CreateBitmap
	Dim Out As OutputStream
	Dim filename As String
	Dim msg As String
	
	filename = pfdfile.ToLowerCase.Replace(".pfd", ".png")
	Try
		Out = File.OpenOutput(File.dirapp, filename, False)
		pfdbitmap.WriteToStream(Out, 100, "PNG")
		Out.Close
		msg = $"[ExportToPNG] PFD exported as PNG image to ${filename}."$
	Catch
		msg = $"[ERROR][ExportToPNG] Can not export PFD as PNG image to ${filename}.${CRLF}${LastException}"$
	End Try
	xui.MsgboxAsync(msg,"Export PFD to PNG Image")
End Sub

'Export all shapes as defined in the external shape definition file pfdcreator.shp.
'The shapes are stored in a subfolder of the dirapp folder.
'The shapename is in lowercase.
'imagefolder - Subfolder of the imagefolder, i.e. images wil strore in file.dirapp/images
Public Sub ExportShapesToPNG(shapes As List, imagefolder As String)
	Try
		File.MakeDir(File.DirApp, imagefolder)
	Catch
		PFDCommon.ShowMessage(PFDCommon.MSG_TYPE_ERROR, _
		                 	  "Export Shapes", _ 
							  $"Can not create folder ${imagefolder}.${CRLF}${LastException}"$)
		Return
	End Try
		
	Try
		'Loop over the shapes and get the type and its shapes
		For Each colshape As Map In shapes
			Dim obj As TPFDObject

			obj.Initialize
			obj.pfdobject		= IIf(colshape.ContainsKey("pfdobject"), colshape.Get("pfdobject"), PFDCommon.OBJECT_UNKNOWN)
			obj.value			= IIf(colshape.ContainsKey("value"), colshape.Get("value"), "")
			obj.width			= IIf(colshape.ContainsKey("width"), colshape.Get("width"), -1)
			obj.height			= IIf(colshape.ContainsKey("height"), colshape.Get("height"), -1)
			obj.direction		= PFDCommon.DIRECTION_NONE
			obj.strokewidth		= IIf(colshape.ContainsKey("strokewidth"), colshape.Get("strokewidth"), -1)
			obj.cornerradius	= IIf(colshape.ContainsKey("cornerradius"), colshape.Get("cornerradius"), -1)
			obj.bordercolor		= IIf(colshape.ContainsKey("bordercolor"), Helper.ColorHexToInt(colshape.Get("bordercolor")), -1)	
			obj.fillcolor		= IIf(colshape.ContainsKey("fillcolor"), Helper.ColorHexToInt(colshape.Get("fillcolor")), -1)		
			obj.textcolor		= IIf(colshape.ContainsKey("textcolor"), Helper.ColorHexToInt(colshape.Get("textcolor")), -1)
			obj.font			= IIf(colshape.ContainsKey("font"), colshape.Get("font"), PFDCommon.FONT_DEFAULT)
			obj.fontsize		= IIf(colshape.ContainsKey("fontsize"), colshape.Get("fontsize"), PFDCommon.FONT_SMALL_SIZE)
			obj.textalignment	= IIf(colshape.ContainsKey("textalignment"), colshape.Get("textalignment"), PFDCommon.TEXT_ALIGNMENT_CENTER)

			'Draw the object to a PNG image
			Dim bmp As B4XBitmap = PFDCommon.DrawObjectToImage(obj)
			Dim out As OutputStream
			out = File.OpenOutput($"${File.DirApp}/${imagefolder}"$, $"${obj.pfdobject.ToLowerCase}.png"$, False)
			bmp.WriteToStream(out, 100, "PNG")
			out.Close
		Next
	Catch
		PFDCommon.ShowMessage(PFDCommon.MSG_TYPE_ERROR, _
		                 	  "Export Shapes", _ 
							  $"Can not export shapes to folder ${imagefolder}.${CRLF}${LastException}"$)
		Return
	End Try

	xui.MsgboxAsync($"Shapes (#${shapes.size}) exported to folder ${imagefolder}."$,"Export Shapes to PNG Images")
End Sub
#End Region

#Region SelectToCode
'Create B4X DrawMethod code from shapes selected (drawmode=select).
'TODO: The code created from the select rectangle, which contains one or more shapes, can not be resized.
'Return - B4X Code DrawShape for the given shapename.
'Example<code>
'If DrawSelectObjects.Size > 0 Then
'	b4xcode = PFD.SelectToCode(shapename, PFDShapes, DrawSelectObjects, DrawSelectObjectsRect)
'End If
'</code>
Public Sub SelectToCode(shapename As String, shapes As List, objects As List, baserect As B4XRect) As String
	Dim result As String
	Dim sb As StringBuilder
	
	'Init the stringbuilder
	sb.Initialize

	'Region and Heading
	sb = Helper.AddCodeLine(sb, False, $"#Region ${shapename}"$)
	sb = Helper.AddCodeLine(sb, False, $"'${shapename}"$)

	'Add info about the select rectangle
	'sb = Helper.AddCodeLine(sb, False, $"'${baserect}"$)
	sb = Helper.AddCodeLine(sb, False, $"'Base Rectangle: width=${baserect.width},height=${baserect.height}"$)
	sb = Helper.AddCodeLine(sb, False, $"'Note: resizing the shape might not work properly - this will require manual code change."$)

	'Add the sub header with sets the shapename in the drawing method with always the same arguments
	sb = Helper.AddCodeLine(sb, False, $"Public Sub Draw${shapename}(cvs As AsyncCanvas, obj As TPFDObject)"$)

	'Sort the object list by index 0 - N
	objects.SortType("index",True)
	
	'Loop over the object and add a map with pfd object properties
	For Each obj As TPFDObject In objects
		'Create an unique object using the object index
		sb = Helper.AddCodeLine(sb, True, $"Dim obj${obj.index} as TPFDObject"$)

		'Init the object
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.initialize"$)
		
		'Set object center x,y adjusted from baserect = baserect to become 0,0,width,height
		Dim cx As Float = baserect.CenterX - obj.centerx
		Dim cy As Float = baserect.CenterY - obj.centery
		
		'Set object width&height factor to handle resizing
		Dim wf As Float = obj.width / baserect.width
		Dim hf As Float = obj.height / baserect.height

		'Set object properties
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.arrow = "${obj.arrow}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.bordercolor = ${obj.bordercolor} '${Helper.ColorIntToHexARGB(obj.bordercolor)}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.centerx = obj.centerx - (${cx})"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.centery = obj.centery - (${cy})"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.cornerradius = ${obj.cornerradius}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.description = "${obj.description}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.direction = "${obj.direction}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.fillcolor = ${obj.fillcolor} '${Helper.ColorIntToHexARGB(obj.fillcolor)}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.font = "${obj.font}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.fontsize = ${obj.fontsize}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.height = obj.height * ${hf} '${obj.height}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.image = "${obj.image}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.operate = ${obj.operate}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.pfdobject = "${obj.pfdobject}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.separator = ${obj.separator}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.strokewidth = ${obj.strokewidth}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.tag = "${obj.tag}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.textalignment = "${obj.textalignment}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.textcolor = ${obj.textcolor} '${Helper.ColorIntToHexARGB(obj.textcolor)}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.unit = "${obj.unit}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.value = "${obj.value}""$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.width = obj.width * ${wf} '${obj.width}"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.x1 = obj${obj.index}.centerx - (obj${obj.index}.width / 2)"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.y1 = obj${obj.index}.centery - (obj${obj.index}.height / 2)"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.x2 = obj${obj.index}.centerx + (obj${obj.index}.width / 2)"$)
		sb = Helper.AddCodeLine(sb, True, $"obj${obj.index}.y2 = obj${obj.index}.centery + (obj${obj.index}.height / 2)"$)

		'Get the method from the shape definitions file pfdcreator.shp
		sb = Helper.AddCodeLine(sb, True, $"PFDShapesBasic.${GetShapeMethod(shapes, obj.pfdobject)}(cvs, obj${obj.index})"$)

		'Empty line as object separator
		sb = Helper.AddCodeLine(sb, False, $""$)
	Next

	'Add the commented code to inform about where cvs commit is performed
	sb = Helper.AddCodeLine(sb, True, $"'Commit the drawing is done in method ReDraw in PFDBase."$)
	sb = Helper.AddCodeLine(sb, True, $"'cvs.Invalidate"$)

	'Add the sub and region end
	sb = Helper.AddCodeLine(sb, False, $"End Sub"$)
	sb = Helper.AddCodeLine(sb, False, $"#End Region"$)

	'Save the code as external file with extension sub
	Try
		File.WriteString(File.DirApp, $"${shapename}.sub"$, sb.ToString)
	Catch
		PFDCommon.showmessage(PFDCommon.MSG_TYPE_ERROR, "Select to Code", $"Can not save the B4X code created.${CRLF}${LastException}"$)
	End Try

	'Return the generated code as string
	result = sb.tostring

	Return result
End Sub
#End Region
