﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=10
@EndOfDesignText@
#Region Code Module Information
'Static Code Module: PFDShapesElectrical
'File: PFDShapesElectrical.bas
'Static code module PFD Object & Shapes Electrical.
'
'PFDCreator Copyright see B4XMainPage.
#End Region

Sub Process_Globals
	'Core
	Private xui As XUI

	'PFD objects specific to this PFD flow type.
	'Each of the PFD objects (UPPERCASE) must match with external shapes file PFDCreator.shp!
	'Common PFD objects are defined in the module PFDShapesBasic.
	'Make Public to access from PFDOperator application..
	Public ELECTRICAL_SHAPE_7SEGMENTDISPLAY As String	= "7SEGMENTDISPLAY"
	Public ELECTRICAL_SHAPE_BATTERY As String			= "BATTERY"
	Public ELECTRICAL_SHAPE_CAPACITOR As String			= "CAPACITOR"
	Public ELECTRICAL_SHAPE_CIRCUITCONNECTOR As String	= "CIRCUITCONNECTOR"
	Public ELECTRICAL_SHAPE_DIODE As String				= "DIODE"
	Public ELECTRICAL_SHAPE_RESISTOR As String			= "RESISTOR"
	Public ELECTRICAL_SHAPE_RESISTORALTERNATE As String	= "RESISTORALTERNATE"
End Sub

#Region DrawShape
'Draw shape(s) selected from pfdobjects list.
'The shapes are defined as process_globals.
Public Sub DrawShapesFromList(cvs As AsyncCanvas, pfdobjects As List)
	'Loop over the pfd objects and draw the object shape
	For Each pfdobject As TPFDObject In pfdobjects
		DrawShape(cvs, pfdobject)
	Next
End Sub

'Draw a shape by shape type as defined in process_globals.
Public Sub DrawShape(cvs As AsyncCanvas, obj As TPFDObject)
	Select obj.pfdobject
		Case ELECTRICAL_SHAPE_7SEGMENTDISPLAY:
			Draw7SegmentDisplay(cvs, obj)
		Case ELECTRICAL_SHAPE_BATTERY:
			DrawBattery(cvs, obj)
		Case ELECTRICAL_SHAPE_CAPACITOR:
			DrawCapacitor(cvs, obj)
		Case ELECTRICAL_SHAPE_CIRCUITCONNECTOR:
			DrawCircuitConnector(cvs, obj)
		Case ELECTRICAL_SHAPE_DIODE:
			DrawDiode(cvs, obj)
		Case ELECTRICAL_SHAPE_RESISTOR:
			DrawResistor(cvs, obj)
		Case ELECTRICAL_SHAPE_RESISTORALTERNATE:
			DrawResistorAlternate(cvs, obj)
	End Select
End Sub
#End Region

#Region 7-Segment-Display
'TM1637 4 digit 7-segment display module
'Draw shape ValueUnit with 2 horizontal lines for value & unit
Public Sub Draw7SegmentDisplay(cvs As AsyncCanvas, obj As TPFDObject)
	Dim cx, cy As Float
	Dim textalignment As String = PFDCommon.TEXT_ALIGNMENT_CENTER
	Dim rect As B4XRect
	Dim x1,y1,x2,y2 As Int

	'Set the font - the font must be digital
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText("A", fnt)

	'Check value
	Dim value As String = obj.value
	If obj.value.length == 0 Then
		value = 0
	else If obj.value.length > 5 Then
		value = "ERR"
	else if Helper.ValueIsNumber(value) Then
		If value.As(Float) < 0 Or value.As(Float) > 9999 Then
			value = "ERR"
		End If
	End If

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText("A", fnt)

	'Set rect dimension
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rect.Initialize(x1,y1,x2,y2)

	'Draw the inner shape
	cvs.DrawRoundRect(rect, obj.fillcolor, True, 0, obj.cornerradius)
	'Draw the outer shape
	cvs.DrawRoundRect(rect, obj.bordercolor, False, obj.strokewidth, obj.cornerradius)

	'Draw value
	cx = obj.centerx
	cy = obj.centery + (fntrect.Height/2)
	cvs.DrawText($"${value}"$, cx, cy, fnt, obj.textcolor, textalignment)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region DrawBattery
'Draw Electrical Shape Battery.
'The shape is drawn using the PFDShapesBasic draw methods with a dedicated object per shape (MANDATORY, do not use same object for a shape).
'There are 4 shapes: hor line + ver line long + ver line short + hor line
'--| +-- split into 5 segments: 2seg (hor line left) + 1seg (gap between hor lines) + 2 seg (hor line right)
'Each segment has a width = object.width / 5.
'The backgroundcolor is the PFD Drawing Board background color.
'Properties: value,centerx,centery,width,height,font,fontsize,textcolor,textalignment.
Public Sub DrawBattery(cvs As AsyncCanvas, obj As TPFDObject)
	Dim cx As Float			= obj.centerx
	Dim cy As Float			= obj.centery
	Dim w As Float			= obj.width
	Dim w2 As Float			= w / 2
	Dim h As Float			= obj.height
	Dim h2 As Float			= h / 2
	Dim SEGMENTS As Int		= 5
	Dim wsegment As Float	= w / SEGMENTS

	'Draw hor line left
	Dim obj1 As TPFDObject
	obj1.Initialize
	obj1.width = wsegment * 2
	obj1.height = h
	obj1.centerx = cx - w2 + wsegment
	obj1.centery = cy
	obj1.strokewidth = obj.strokewidth
	obj1.fillcolor = obj.fillcolor
	obj1.direction = PFDCommon.DIRECTION_HORIZONTAL
	PFDShapesBasic.DrawLineSolid(cvs, obj1)

	'Draw ver line left
	Dim obj2 As TPFDObject
	obj2.Initialize
	obj2.width = obj.strokewidth
	obj2.height = h
	obj2.centerx = obj1.centerx + wsegment
	obj2.centery = cy
	obj2.strokewidth = obj.strokewidth
	obj2.fillcolor = obj.fillcolor
	obj2.direction = PFDCommon.DIRECTION_VERTICAL
	PFDShapesBasic.DrawLineSolid(cvs, obj2)

	'Draw ver line right
	Dim obj3 As TPFDObject
	obj3.Initialize
	obj3.width = obj.strokewidth
	obj3.height = h2
	obj3.centerx = obj2.centerx + wsegment
	obj3.centery = cy
	obj3.strokewidth = obj.strokewidth
	obj3.fillcolor = obj.fillcolor
	obj3.direction = PFDCommon.DIRECTION_VERTICAL
	PFDShapesBasic.DrawLineSolid(cvs, obj3)

	'Draw hor line right
	Dim obj4 As TPFDObject
	obj4.Initialize
	obj4.width = wsegment * 2
	obj4.height = h
	obj4.centerx = obj3.centerx + wsegment
	obj4.centery = cy
	obj4.strokewidth = obj.strokewidth
	obj4.fillcolor = obj.fillcolor
	obj4.direction = PFDCommon.DIRECTION_HORIZONTAL
	PFDShapesBasic.DrawLineSolid(cvs, obj4)
End Sub
#End Region

#Region DrawConnector
'Draw a connector which is a filled circle.
Public Sub DrawCircuitConnector(cvs As AsyncCanvas, obj As TPFDObject)
	cvs.DrawCircle(obj.centerx, obj.centery, obj.width/2, obj.fillcolor, True, 0)
End Sub
#End Region

#Region DrawCapacitor
'Draw shape Capacitor with connector lines and text.
'The resistor has additional text right or below the shape.
'These are 2 lines with Tag and ValueUnit. Ensure to set the textalignment to center for hor and to left for vert.
'The unit can contain unicode characters, like Ohm sign \u2126.
'--||--
' TAG
'VALUEUNIT
'The unicode for the mu character is \u00B5.
Public Sub DrawCapacitor(cvs As AsyncCanvas, obj As TPFDObject)
	Dim SEGMENTS As Int = 10	'The resistor is split into 10 segments used to draw the lines
	Dim segmentsize As Float	'Size of each segment to draw a line
	Dim x1,y1,x2,y2 As Float	'Line start and end pos
	Dim points As List			'List of points x1,y1,x2,y2 to draw the lines
	Dim textlines As List		'Two text lines for tag + valueunit
	Dim w As Float= obj.width	'Object width
	Dim h As Float= obj.height	'Object height
	Dim w2 As Float= w / 2		'Object halfwidth
	Dim h2 As Float= h / 2		'Object halfheight
	
	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrecttag As B4XRect = cvs.MeasureText(obj.tag, fnt)
	Dim TEXTOFFSET As Int = 2dip

	'Init the list of points and the textlines (for tag + valueunit)
	points.Initialize
	textlines.Initialize

	'Adjust obj w&h to fit within shape selector
	If obj.shapeselector == obj.pfdobject Then
		w = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		w2 = w / 2
		h2 = h / 2
	End If
		
	'Check the direction and set to hor if not hor or vert
	If Not(Helper.ItemInArray(obj.direction, Array As String(PFDCommon.DIRECTION_HORIZONTAL, PFDCommon.DIRECTION_VERTICAL))) Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If
	
	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		'Draw the capacitor hor with tag+value+unit centered below the object
		'Set obj.textalignment to CENTER

		'Set the segmentsize using the width
		segmentsize = w / SEGMENTS

		'S1-4 = Left hor line + vert line
		x1 = obj.centerx - w2
		y1 = obj.centery
		x2 = x1 + (segmentsize * 4.5)
		y2 = y1
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		x1 = x2
		y1 = y2 - h2
		x2 = x1
		y2 = y1 + h
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		'S5-6 - Space

		'S7-10 = Right vert line + hor line
		x1 = x2 + (segmentsize * 1)
		y1 = obj.centery - h2
		x2 = x1
		y2 = y1 + h
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)
		x1 = x2
		y1 = obj.centery
		x2 = x1 + (segmentsize * 4.5)
		y2 = y1
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx
		y1 = obj.centery + h2 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
	End If

	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		'Draw the resistor vert with tag+value+unit left alinged right from the object
		'Set obj.textalignment to LEFT

		'Set the segmentsize using the height
		segmentsize = h / SEGMENTS

		'S1-4 = Mid vert line + hor line
		x1 = obj.centerx
		y1 = obj.centery - h2
		x2 = x1
		y2 = y1 + (segmentsize * 4.5)
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		x1 = x2 - w2
		y1 = y2
		x2 = x1 + w
		y2 = y1
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		'S5-6 - Space

		'S7-10 = Right hor line + vert line
		x1 = obj.centerx - w2
		y1 = y2 + (segmentsize * 1)
		x2 = x1 + w
		y2 = y1
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)

		x1 = obj.centerx
		y1 = y2
		x2 = x1
		y2 = y1 + (segmentsize * 4.5)
		cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor,obj.strokewidth)
		
		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx + w2 + TEXTOFFSET
		y1 = obj.centery
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
	End If
	
	'Draw the lines
	For i = 0 To points.size - 1
		Dim p() As Float = points.Get(i)
		cvs.DrawLine(p(0),p(1),p(2),p(3),obj.bordercolor,obj.strokewidth)
	Next

	'Check if not drawing the shapeselecter = do not display text
	If Not(obj.shapeselector.EqualsIgnoreCase(obj.pfdobject)) Then
		'Draw the twp textlines tag, value and unit always horizontal
		'Tag
		Dim l() As Float = textlines.Get(0)
		cvs.DrawText($"${obj.tag}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
		'Value + Unit
		Dim l() As Float = textlines.Get(1)
		cvs.DrawText($"${obj.value}${Helper.DecodeUnicodeString(obj.unit)}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region DrawDiode
'Draw shape Diode which can have 4 directions left, up, down, right.
'The shape is drawn initially with direction right.
'|\
'|+|
'|/
Public Sub DrawDiode(cvs As AsyncCanvas, obj As TPFDObject)
	Dim path As B4XPath
	Dim x, y As Float
	Dim w As Float= obj.width	'Object width
	Dim h As Float= obj.height	'Object height
	Dim w2 As Float= w / 2		'Object halfwidth
	Dim h2 As Float= h / 2		'Object halfheight
	Dim sw2 As Float = obj.strokewidth / 2
	Dim degrees As Float

	'Adjust obj w&h to fit within shape selector
	If obj.shapeselector == obj.pfdobject Then
		w = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		w = PFDCommon.SHAPE_SELECTOR_WIDTH / 2
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT / 2
		w2 = w / 2
		h2 = h / 2
		sw2 = 2
	End If

	'Select the direction, set the angle degrees and adjust cx or cy.
	Select obj.direction
		Case PFDCommon.DIRECTION_RIGHT
			degrees = 0
		Case PFDCommon.DIRECTION_LEFT
			degrees = 180
		Case PFDCommon.DIRECTION_DOWN
			degrees = 90
		Case PFDCommon.DIRECTION_UP
			degrees = -90
	End Select

	'Define the path for a shape direction right
	'MidRight
	x = obj.centerx + w2
	y = obj.centery
	path.Initialize(x, y)

	'TopLeft
	x = obj.centerx - w2
	y = obj.centery - h2
	path.LineTo(x, y)

	'BotLeft
	x = obj.centerx - w2
	y = obj.centery + h2
	path.LineTo(x, y)

	'MidRight
	x = obj.centerx + w2
	y = obj.centery
	path.LineTo(x, y)

	'TopRightR
	x = obj.centerx + w2 + sw2
	y = obj.centery - h2
	path.LineTo(x, y)

	'BotRightR
	x = obj.centerx + w2 + sw2
	y = obj.centery + h2
	path.LineTo(x, y)

	'BotRightL
	x = obj.centerx + w2 - sw2
	y = obj.centery + h2
	path.LineTo(x, y)

	'TopRightL
	x = obj.centerx + w2 - sw2
	y = obj.centery - h2
	path.LineTo(x, y)

	'TopRightR
	x = obj.centerx + w2 + sw2
	y = obj.centery - h2
	path.LineTo(x, y)

	'Draw the inner shape
	cvs.DrawPathRotated(path, obj.fillcolor, True, 0, degrees, obj.centerx, obj.centery)
	'Draw the outer shape
	cvs.DrawPathRotated(path, obj.bordercolor, False, obj.strokewidth, degrees, obj.centerx, obj.centery)
End Sub
#End Region

#Region DrawResistor
'Draw shape Resistor hor or vert as zigzag shape with start and end connector lines.
'The resistor has additional text right or below the shape.
'These are 2 lines with Tag and ValueUnit. Ensure to set the textalignment to center for hor and to left for vert.
'       3     5     7
'0-1   / \   / \   / \   9-10
'   \ /   \ /   \ /   \ /
'    2     4     6     8
'S1S2S3  S4 S5 S6 S7 S8 S9S10  
'    TAG
'   VALUEUNIT
Public Sub DrawResistor(cvs As AsyncCanvas, obj As TPFDObject)
	Dim SEGMENTS As Int = 10	'The resistor is split into 10 segments used to draw the lines
	Dim segmentsize As Float	'Size of each segment to draw a line
	Dim x1,y1,x2,y2 As Float	'Line start and end pos
	Dim points As List			'List of points x1,y1,x2,y2 to draw the lines
	Dim textlines As List		'Two text lines for tag + valueunit
	Dim w As Float= obj.width	'Object width
	Dim h As Float= obj.height	'Object height
	Dim w2 As Float= w / 2		'Object halfwidth
	Dim h2 As Float= h / 2		'Object halfheight
	
	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrecttag As B4XRect = cvs.MeasureText(obj.tag, fnt)
	Dim TEXTOFFSET As Int = 4dip

	'Init the list of points and the textlines (for tag + valueunit)
	points.Initialize
	textlines.Initialize

	'Adjust obj w&h to fit within shape selector
	If obj.shapeselector == obj.pfdobject Then
		w = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		w2 = w / 2
		h2 = h / 2
	End If
	
	'Check the direction and set to hor if not hor or vert
	If Not(Helper.ItemInArray(obj.direction, Array As String(PFDCommon.DIRECTION_HORIZONTAL, PFDCommon.DIRECTION_VERTICAL))) Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If
	
	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		'Draw the resistor hor with tag+value+unit centered below the object
		'Set obj.textalignment to CENTER

		'Set the segmentsize using the width
		segmentsize = w / SEGMENTS

		'S1 = left hor line plus zig down
		x1 = obj.centerx - w2
		y1 = obj.centery
		x2 = x1 + segmentsize
		y2 = y1
		points.Add(Array As Float(x1,y1,x2,y2))
		'S2 = zig right down
		x1 = x2
		y1 = y2
		x2 = x1 + segmentsize
		y2 = y2 + h2
		points.Add(Array As Float(x1,y1,x2,y2))

		'Drawing the 3 zigzags for the segments 3-8
		For i = 1 To 3
			'S = zig right up
			x1 = x2
			y1 = y2
			x2 = x1 + segmentsize
			y2 = y1 - h
			points.Add(Array As Float(x1,y1,x2,y2))
			'S = zig right down
			x1 = x2
			y1 = y2
			x2 = x1 + segmentsize
			y2 = y1 + h
			points.Add(Array As Float(x1,y1,x2,y2))
		Next

		'Right hor line plus zig up
		'S9 = zig right up
		x1 = x2
		y1 = y2
		x2 = x1 + segmentsize
		y2 = obj.centery
		points.Add(Array As Float(x1,y1,x2,y2))
		'S10 = hor line
		x1 = x2
		y1 = y2
		x2 = x1 + segmentsize
		y2 = y2
		points.Add(Array As Float(x1,y1,x2,y2))
		
		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx
		y1 = obj.centery + h2 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
	End If

	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		'Draw the resistor vert with tag+value+unit left alinged right from the object
		'Set obj.textalignment to LEFT

		'Set the segmentsize using the height
		segmentsize = h / SEGMENTS

		'TopMid vert line plus zig left down
		'S1 = vert line
		x1 = obj.centerx
		y1 = obj.centery - h2
		x2 = x1
		y2 = y1 + (segmentsize)
		points.Add(Array As Float(x1,y1,x2,y2))
		'S2 = zig left down
		x1 = x2
		y1 = y2
		x2 = x2 - w2
		y2 = y2 + segmentsize
		points.Add(Array As Float(x1,y1,x2,y2))

		'Drawing the 3 zigzags for the segments 3-8
		For i = 1 To 3
			'S zig right down
			x1 = x2
			y1 = y2
			x2 = x1 + w
			y2 = y1 + segmentsize
			points.Add(Array As Float(x1,y1,x2,y2))
			'S zig left down
			x1 = x2
			y1 = y2
			x2 = x1 - w
			y2 = y1 + segmentsize
			points.Add(Array As Float(x1,y1,x2,y2))
		Next

		'Botmid vert line plus zig right down
		'S9 - zig right down
		x1 = x2
		y1 = y2
		x2 = x1 + w2
		y2 = y1 + segmentsize
		points.Add(Array As Float(x1,y1,x2,y2))
		'S10 - hor line down
		x1 = x2
		y1 = y2
		x2 = x1
		y2 = y2 + segmentsize
		points.Add(Array As Float(x1,y1,x2,y2))

		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx + w2 + TEXTOFFSET
		y1 = obj.centery - (fntrecttag.Height / 2)
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height * 2)
		textlines.Add(Array As Float(x1, y1))
	End If
	
	'Draw the lines
	For i = 0 To points.size - 1
		Dim p() As Float = points.Get(i)
		cvs.DrawLine(p(0),p(1),p(2),p(3),obj.fillcolor,obj.strokewidth)
	Next

	'Check if not drawing the shapeselecter = do not display text
	If Not(obj.shapeselector.EqualsIgnoreCase(obj.pfdobject)) Then
		'Draw the two textlines tag, value and unit always horizontal
		'Tag
		Dim l() As Float = textlines.Get(0)
		cvs.DrawText($"${obj.tag}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
		'Value + Unit
		Dim l() As Float = textlines.Get(1)
		cvs.DrawText($"${obj.value}${Helper.DecodeUnicodeString(obj.unit)}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region DrawResistorAlternate
'Draw shape Resistor as Rectangle with connector lines and text.
'The resistor has additional text right or below the shape.
'These are 2 lines with Tag and ValueUnit. Ensure to set the textalignment to center for hor and to left for vert.
'The unit can contain unicode characters, like Ohm sign \u2126.
'  +-----+
'--|     |--
'  +-----+
'    TAG
'   VALUEUNIT
Public Sub DrawResistorAlternate(cvs As AsyncCanvas, obj As TPFDObject)
	Dim SEGMENTS As Int = 10	'The resistor is split into 10 segments used to draw the lines
	Dim segmentsize As Float	'Size of each segment to draw a line
	Dim x1,y1,x2,y2 As Float	'Line start and end pos
	Dim points As List			'List of points x1,y1,x2,y2 to draw the lines
	Dim textlines As List		'Two text lines for tag + valueunit
	Dim w As Float= obj.width	'Object width
	Dim h As Float= obj.height	'Object height
	Dim w2 As Float= w / 2		'Object halfwidth
	Dim h2 As Float= h / 2		'Object halfheight
	
	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrecttag As B4XRect = cvs.MeasureText(obj.tag, fnt)
	Dim TEXTOFFSET As Int = 2dip

	'Init the list of points and the textlines (for tag + valueunit)
	points.Initialize
	textlines.Initialize

	'Adjust obj w&h to fit within shape selector
	If obj.shapeselector == obj.pfdobject Then
		w = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		w2 = w / 2
		h2 = h / 2
	End If
	
	'Check the direction and set to hor if not hor or vert
	If Not(Helper.ItemInArray(obj.direction, Array As String(PFDCommon.DIRECTION_HORIZONTAL, PFDCommon.DIRECTION_VERTICAL))) Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If
	
	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		'Draw the resistor hor with tag+value+unit centered below the object
		'Set obj.textalignment to CENTER

		'Set the segmentsize using the width
		segmentsize = w / SEGMENTS

		'S1 = Left hor line
		x1 = obj.centerx - w2
		y1 = obj.centery
		x2 = x1 + (segmentsize * 2)
		y2 = y1
		points.Add(Array As Float(x1,y1,x2,y2))

		'Draw rectangle for the segments 3-8
		Dim rect As B4XRect
		'TopLeft
		x1 = x2
		y1 = y2 - h2 + (obj.strokewidth / 2)
		'BotRight
		x2 = x1 + (segmentsize * 6)
		y2 = y1 + h - (obj.strokewidth)
		rect.Initialize(x1,y1,x2,y2)
		cvs.DrawRect(rect, obj.fillcolor, True, obj.strokewidth)
		cvs.DrawRect(rect, obj.bordercolor, False, obj.strokewidth)

		'Right hor line
		'S10 - hor line
		x1 = obj.centerx + w2 - (segmentsize * 2)
		y1 = obj.centery
		x2 = x1 + (segmentsize * 2)
		y2 = y1
		points.Add(Array As Float(x1,y1,x2,y2))
		
		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx
		y1 = obj.centery + h2 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
	End If

	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		'Draw the resistor vert with tag+value+unit left alinged right from the object
		'Set obj.textalignment to LEFT

		'Set the segmentsize using the height
		segmentsize = h / SEGMENTS

		'S1 = Top vert line
		x1 = obj.centerx
		y1 = obj.centery - h2
		x2 = x1
		y2 = y1 + (segmentsize * 2)
		points.Add(Array As Float(x1,y1,x2,y2))

		'Draw rectangle for the segments 3-8
		Dim rect As B4XRect
		'TopLeft
		x1 = x2 - w2 + (obj.strokewidth / 2)
		y1 = y2
		'BotRight
		x2 = x1 + w - obj.strokewidth
		y2 = y1 + (segmentsize * 6)
		rect.Initialize(x1,y1,x2,y2)
		cvs.DrawRect(rect, obj.fillcolor, True, obj.strokewidth)
		cvs.DrawRect(rect, obj.bordercolor, False, obj.strokewidth)

		'Bot vert line
		'S10 - vert line
		x1 = obj.centerx
		y1 = obj.centery + h2 - (segmentsize * 2)
		x2 = x1
		y2 = y1 + (segmentsize *2)
		points.Add(Array As Float(x1,y1,x2,y2))
		
		'Set the text lines pos for the tag + valueunit
		x1 = obj.centerx + w2 + TEXTOFFSET
		y1 = obj.centery
		textlines.Add(Array As Float(x1, y1))
		x1 = x1
		y1 = y1 + (fntrecttag.Height) + TEXTOFFSET
		textlines.Add(Array As Float(x1, y1))
	End If
	
	'Draw the lines
	For i = 0 To points.size - 1
		Dim p() As Float = points.Get(i)
		cvs.DrawLine(p(0),p(1),p(2),p(3),obj.bordercolor,obj.strokewidth)
	Next

	'Check if not drawing the shapeselecter = do not display text
	If Not(obj.shapeselector.EqualsIgnoreCase(obj.pfdobject)) Then
		'Draw the twp textlines tag, value and unit always horizontal
		'Tag
		Dim l() As Float = textlines.Get(0)
		cvs.DrawText($"${obj.tag}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
		'Value + Unit
		Dim l() As Float = textlines.Get(1)
		cvs.DrawText($"${obj.value}${Helper.DecodeUnicodeString(obj.unit)}"$, l(0), l(1), fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

