﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=10
@EndOfDesignText@
#Region Code Module Information
'Static Code Module: PFDShapesFlowchart
'File: PFDShapesFlowchart.bas
'Static code module for the PFD Flowchart Object & Shapes.
'The flowchart objects text is set from the object value.
'The tag is not used for display purposes.
'
'PFDCreator Copyright see B4XMainPage.
#End Region

Sub Process_Globals
	'Core
	Private xui As XUI

	'PFD objects specific to this PFD flow type.
	'Each of the PFD objects (UPPERCASE) must match with external shapes file PFDFlowchart.shp!
	'Common PFD objects are defined in the module PFDShapesBasic.
	'Make Public to access from PFDOperator application..
	Public FLOWCHART_SHAPE_COLLATE As String			= "COLLATE"
	Public FLOWCHART_SHAPE_CONNECTOR As String			= "CONNECTOR"
	Public FLOWCHART_SHAPE_DATA As String				= "DATA"
	Public FLOWCHART_SHAPE_DATASTORAGE As String		= "DATASTORAGE"
	Public FLOWCHART_SHAPE_DATABASE As String			= "DATABASE"
	Public FLOWCHART_SHAPE_DECISION As String			= "DECISION"
	Public FLOWCHART_SHAPE_DELAY As String				= "DELAY"
	Public FLOWCHART_SHAPE_DISPLAY As String			= "DISPLAY"
	Public FLOWCHART_SHAPE_DOCUMENT As String			= "DOCUMENT"
	Public FLOWCHART_SHAPE_INTERNALSTORAGE As String	= "INTERNALSTORAGE"
	Public FLOWCHART_SHAPE_JUNCTION As String			= "JUNCTION"
	Public FLOWCHART_SHAPE_LOOPLIMIT As String			= "LOOPLIMIT"
	Public FLOWCHART_SHAPE_MANUALINPUT As String		= "MANUALINPUT"
	Public FLOWCHART_SHAPE_MANUALLOOP As String			= "MANUALLOOP"
	Public FLOWCHART_SHAPE_MERGE As String				= "MERGE"
	Public FLOWCHART_SHAPE_OFFPAGE As String			= "OFFPAGE"
	Public FLOWCHART_SHAPE_OPERATION As String			= "OPERATION"
	Public FLOWCHART_SHAPE_OR As String					= "OR"
	Public FLOWCHART_SHAPE_PREPARATION As String		= "PREPARATION"
	Public FLOWCHART_SHAPE_PROCESS As String			= "PROCESS"
	Public FLOWCHART_SHAPE_SORT As String				= "SORT"
	Public FLOWCHART_SHAPE_SUBROUTINE As String			= "SUBROUTINE"
	Public FLOWCHART_SHAPE_TERMINATOR As String			= "TERMINATOR"
End Sub

#Region DrawShape
'Draw shape(s) selected from pfdobjects list.
'The shapes are defined as process_globals.
Public Sub DrawShapesFromList(cvs As AsyncCanvas, pfdobjects As List)
	'Loop over the pfd objects and draw the object shape
	For Each pfdobject As TPFDObject In pfdobjects
		DrawShape(cvs, pfdobject)
	Next
End Sub

'Draw a shape by shape type as defined in process_globals.
Public Sub DrawShape(cvs As AsyncCanvas, obj As TPFDObject)
	Select obj.pfdobject
		Case FLOWCHART_SHAPE_COLLATE:
			DrawCollate(cvs, obj)
		Case FLOWCHART_SHAPE_CONNECTOR:
			DrawConnector(cvs, obj)
		Case FLOWCHART_SHAPE_DATA:
			DrawData(cvs, obj)
		Case FLOWCHART_SHAPE_DATABASE:
			DrawDatabase(cvs, obj)
		Case FLOWCHART_SHAPE_DATASTORAGE:
			DrawDataStorage(cvs, obj)
		Case FLOWCHART_SHAPE_DELAY:
			DrawDelay(cvs, obj)
		Case FLOWCHART_SHAPE_DECISION:
			DrawDecision(cvs, obj)
		Case FLOWCHART_SHAPE_DISPLAY:
			DrawDisplay(cvs, obj)
		Case FLOWCHART_SHAPE_DOCUMENT:
			DrawDocument(cvs, obj)
		Case FLOWCHART_SHAPE_LOOPLIMIT:
			DrawLoopLimit(cvs, obj)
		Case FLOWCHART_SHAPE_MANUALINPUT:
			DrawManualInput(cvs, obj)
		Case FLOWCHART_SHAPE_MANUALLOOP:
			DrawManualLoop(cvs, obj)
		Case FLOWCHART_SHAPE_MERGE:
			DrawMerge(cvs, obj)
		Case FLOWCHART_SHAPE_INTERNALSTORAGE:
			DrawInternalStorage(cvs, obj)
		Case FLOWCHART_SHAPE_JUNCTION:
			DrawJunction(cvs, obj)
		Case FLOWCHART_SHAPE_OFFPAGE:
			DrawOffPage(cvs, obj)
		Case FLOWCHART_SHAPE_OPERATION:
			DrawOperation(cvs, obj)
		Case FLOWCHART_SHAPE_OR:
			DrawOR(cvs, obj)
		Case FLOWCHART_SHAPE_PREPARATION:
			DrawPreparation(cvs, obj)
		Case FLOWCHART_SHAPE_PROCESS:
			DrawProcess(cvs, obj)
		Case FLOWCHART_SHAPE_SORT:
			DrawSort(cvs, obj)
		Case FLOWCHART_SHAPE_SUBROUTINE:
			DrawSubRoutine(cvs, obj)
		Case FLOWCHART_SHAPE_TERMINATOR:
			DrawTerminator(cvs, obj)
	End Select
End Sub
#End Region

#Region Objects
'Draw flowchart collate shape.
Public Sub DrawCollate(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath
	
	'TopLeft
	p.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'BotLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))
	'BotRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	'TopLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

#Region Connector
'Draw flowchart connector shape.
Public Sub DrawConnector(cvs As AsyncCanvas, obj As TPFDObject)
	'Get the width adjusted with the strokewidth
	Dim w As Int = obj.width - obj.strokewidth
	
	'Draw the inner shape
	cvs.DrawCircle(obj.centerx, obj.centery, (w/2), obj.fillcolor, True, 0)
	'Draw the outer shape
	If obj.strokewidth > 0 Then cvs.DrawCircle(obj.centerx, obj.centery, (w/2), obj.bordercolor, False, obj.strokewidth)

	'Show the value centered in the object
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Data
'Draw flowchart data.
Public Sub DrawData(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopRight
	p.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to Topleft
	p.LineTo(obj.centerx - (obj.width/4), obj.centery - (obj.height/2))
	'TopLeft to BottomLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))
	'BottomLeft to BottomRight
	p.LineTo(obj.centerx + (obj.width/4), obj.centery + (obj.height/2))
	'BottomRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region DataStorage
'Draw flowchart data storage shape.
Public Sub DrawDataStorage(cvs As AsyncCanvas, obj As TPFDObject)
	' Initialize the canvas and path
	' 20% of the height for the arc radius
	Dim ArcRadius As Int = 0.3 * obj.height
	
	Dim pathl As B4XPath
	'TopLeft
	pathl.Initialize(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery - (obj.height/2))
	'TopLeft to BotLeft
	pathl.LineTo(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery + (obj.height/2))
	'BotLeft to BotRight
	pathl.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	'BotRight to TopRight
	pathl.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to TopLeft
	pathl.LineTo(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery - (obj.height/2))
	' Draw the path on the canvas
	cvs.DrawPath(pathl, obj.Fillcolor, True, 0)
	cvs.DrawPath(pathl, obj.bordercolor, False, obj.strokewidth)

	'Left Oval (centered horizontally)
	Dim pathot As B4XPath
	Dim rectot As B4XRect
	rectot.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2), _
					  obj.centerx - (obj.width/2) + ArcRadius, obj.centery + (obj.height/2))
	pathot.InitializeOval(rectot)
	cvs.DrawPath(pathot, obj.fillcolor, True, 0)
	cvs.DrawPath(pathot, obj.bordercolor, False, obj.strokewidth)

	Dim pathi As B4XPath
	'TopLeft
	pathi.Initialize(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery - (obj.height/2) + obj.strokewidth)
	'TopLeft to BotLeft
	pathi.LineTo(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery + (obj.height/2) - obj.strokewidth)
	'BotLeft to BotRight
	pathi.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2) - obj.strokewidth)
	'BotRight to TopRight
	pathi.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2) + obj.strokewidth)
	'TopRight to TopLeft
	pathi.LineTo(obj.centerx - (obj.width/2) + (ArcRadius/2), obj.centery - (obj.height/2) + obj.strokewidth)
	' Draw the path on the canvas
	cvs.DrawPath(pathi, obj.Fillcolor, True, 0)

	'RightOval (centered horizontally)
	Dim pathob As B4XPath
	Dim rectob As B4XRect
	rectob.Initialize(obj.centerx + (obj.width/2) - (ArcRadius/2), obj.centery - (obj.height/2), _
					  obj.centerx + (obj.width/2) + (ArcRadius/2), obj.centery + (obj.height/2))
	pathob.InitializeOval(rectob)
	cvs.DrawPath(pathob, obj.fillcolor, True, 0)
	cvs.DrawPath(pathob, obj.bordercolor, False, obj.strokewidth)

	Dim rector As B4XRect
	rector.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2), _
					  obj.centerx + (obj.width/2) + (ArcRadius/2), obj.centery + (obj.height/2))
	cvs.DrawRect(rector, obj.fillcolor, True, 0)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

Public Sub DrawDatabase(cvs As AsyncCanvas, obj As TPFDObject)
	'20% of the height for the arc radius
	Dim ArcRadius As Int = 0.2 * obj.height
	
	Dim pathl As B4XPath
	'TopLeft
	pathl.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2) + ArcRadius/2)
	'TopLeft to BotLeft
	pathl.Initialize(obj.centerx - (obj.width/2), obj.centery + (obj.height/2) - ArcRadius/2)
	'BotLeft to BotRight
	pathl.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2) - ArcRadius/2)
	'BotRight to TopRight
	pathl.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2) + ArcRadius/2)
	'TopRight to TopLeft
	pathl.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2) + ArcRadius/2)
	' Draw the path on the canvas
	cvs.DrawPath(pathl, obj.Fillcolor, True, 0)
	cvs.DrawPath(pathl, obj.bordercolor, False, obj.strokewidth)

	'Top Oval (centered horizontally)
	Dim pathot As B4XPath
	Dim rectot As B4XRect
	rectot.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2), _
					  obj.centerx + (obj.width/2), obj.centery - (obj.height/2) + ArcRadius)
	pathot.InitializeOval(rectot)
	cvs.DrawPath(pathot, obj.fillcolor, True, 0)
	cvs.DrawPath(pathot, obj.bordercolor, False, obj.strokewidth)

	'BotOval (centered horizontally)
	Dim pathob As B4XPath
	Dim rectob As B4XRect
	rectob.Initialize(obj.centerx - (obj.width/2), obj.centery + (obj.height/2 - ArcRadius), _
					  obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	pathob.InitializeOval(rectob)
	cvs.DrawPath(pathob, obj.fillcolor, True, 0)
	cvs.DrawPath(pathob, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart decision.
'If the length of the value > decision width - 25% then draw the text with crlf in small font.
Public Sub DrawDecision(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'Top
	p.Initialize(obj.centerx, obj.centery - (obj.height/2))
	'Top to left
	p.LineTo(obj.centerx - (obj.width/2), obj.centery)
	'Left to bottom
	p.LineTo(obj.centerx, obj.centery + (obj.height/2))
	'Bottom to right
	p.LineTo(obj.centerx + (obj.width/2), obj.centery)
	'Right to top
	p.LineTo(obj.centerx, obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart delay.
Public Sub DrawDelay(cvs As AsyncCanvas, obj As TPFDObject)
	Dim Path As B4XPath
	Dim x1, y1 As Int
	Dim w As Int
	Dim r As Float = obj.height/2
	
	'Width topline
	w = obj.width - r
	
	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	Path.Initialize(x1, y1)

	'TopRight
	x1 = obj.centerx - (obj.width/2) + w + r	'(obj.width/2) - (w)
	y1 = obj.centery - (obj.height/2)
	Path.LineTo(x1, y1)

	'TopRight to BotRight Arc (=shape with rounded right side)
	Dim i As Float
	Dim StepAngle As Float = 5 ' Smaller step angle gives smoother arc
	'Approximate the top-right corner (rounded) with small line segments
	'Note the angle 0 is clock 0300, -90 = clock 1200, 90 = clock 0600
	For i = -90 To 90 Step StepAngle
		Dim x As Float = x1 - r + r * CosD(i)
		Dim y As Float = y1 + r + r * SinD(i)
		Path.LineTo(x, y)
	Next

	'BotRight to BotLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery + (obj.height/2)
	Path.LineTo(x1, y1)

	'BotLeft to TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	Path.LineTo(x1, y1)

	'Draw the inner shape
	cvs.DrawPath(Path, obj.fillcolor, True, obj.StrokeWidth)
	'Draw the outer shape
	If obj.strokewidth > 0 Then cvs.DrawPath(Path, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart display.
Public Sub DrawDisplay(cvs As AsyncCanvas, obj As TPFDObject)
	Dim path As B4XPath
	Dim x1, y1 As Int

	'Leftside arc radius
	Dim r As Float = obj.height/2

	'MidRight
	x1 = obj.centerx + (obj.width/2)
	y1 = obj.centery
	path.Initialize(x1, y1)

	'MidRight to TopRight
	x1 = obj.centerx + (obj.width/2) - r
	y1 = obj.centery - (obj.height/2)
	path.LineTo(x1, y1)

	'TopRight to TopLeft
	x1 = obj.centerx - (obj.width/2) + (r * 2)
	y1 = obj.centery - (obj.height/2)
	path.LineTo(x1, y1)

	'TopLeft to BotLeft Arc (=shape with rounded right side)
	Dim i As Float
	Dim StepAngle As Float = -5 ' Smaller step angle gives smoother arc
	'Approximate the top-left corner (rounded) with small line segments
	'Note the angle 0 is clock 0300, 270 = clock 1200, 90 = clock 0600
	For i = 270 To 90 Step StepAngle
		Dim x As Float = x1 - r + r * CosD(i)
		Dim y As Float = y1 + r + r * SinD(i)
		path.LineTo(x, y)
	Next

	'BotLeft to BotRight
	x1 = obj.centerx + (obj.width/2) - r
	y1 = obj.centery + (obj.height/2)
	path.LineTo(x1, y1)

	'BotRight to MidRight
	x1 = obj.centerx + (obj.width/2)
	y1 = obj.centery
	path.LineTo(x1, y1)

	'Draw the inner shape
	cvs.DrawPath(path, obj.fillcolor, True, obj.StrokeWidth)
	'Draw the outer shape
	If obj.strokewidth > 0 Then cvs.DrawPath(path, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart document.
'Note the shape has at the bottomm a bezier curve with 2 control points.
Public Sub DrawDocument(cvs As AsyncCanvas, obj As TPFDObject)
	Dim controlpoint_offset As Int = 20		'Control points 1 (above) & 2 (below) baseline
	Dim rectheight_factor As Float = 0.95	'Height size is 0.95 width size
	
	Dim w As Float = obj.width		'Initial width
	Dim h As Float = obj.height		'Initial height

	'Check if shape to be drawn for the shapeselector
	If obj.shapeselector == obj.pfdobject Then
		w = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		h = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR * 0.65
		controlpoint_offset = 10
	End If

	' Define the dimensions of the rectangle
	Dim rectLeft As Float = obj.centerx - (w/2)
	Dim rectTop As Float = obj.centery - (h/2)
	Dim rectWidth As Float = w
	Dim rectheight As Float = h * rectheight_factor

	Dim rectRight As Float = rectLeft + rectWidth
	Dim rectBottom As Float = rectTop + rectheight

	Dim Path As B4XPath
	Path.Initialize(rectLeft, rectTop)

	' Define the control points and end points for the Bézier curve at the bottom (left to right)
	Dim startX As Float = rectLeft 	'- rectWidth
	Dim startY As Float = rectBottom '- rectHeight

	' Control points adjusted for better curve influence
	Dim controlX1 As Float = rectLeft + rectWidth / 3
	Dim controlY1 As Float = rectBottom - controlpoint_offset    ' Control point 1 above the baseline

	Dim controlX2 As Float = rectLeft + 2 * rectWidth / 3
	Dim controlY2 As Float = rectBottom + controlpoint_offset    ' Control point 2 below the baseline

	Dim endX As Float = rectRight
	Dim endY As Float = rectBottom

	' Debug: Use fewer steps to make the process more visible
	Dim steps As Int = 100  ' Higher steps for smoother curve
	Dim t As Float = 0
	Dim increment As Float = 1 / steps

	Path.LineTo(startX, startY)

	' Loop through each step and draw the approximate Bézier curve using a Do While loop
	Do While t <= 1
		' Calculate the next point on the curve using the cubic Bézier formula
		Dim t1 As Float = 1 - t
		Dim t1Sq As Float = t1 * t1
		Dim t1Cb As Float = t1Sq * t1
		Dim tSq As Float = t * t
		Dim tCb As Float = tSq * t
        
		' Calculate current point on Bézier curve
		Dim currentX As Float = t1Cb * startX + 3 * t1Sq * t * controlX1 + 3 * t1 * tSq * controlX2 + tCb * endX
		Dim currentY As Float = t1Cb * startY + 3 * t1Sq * t * controlY1 + 3 * t1 * tSq * controlY2 + tCb * endY

		' Draw a line segment from the last point to the current point
		Path.LineTo(currentX, currentY)

		' Increment t
		t = t + increment
	Loop
	Path.LineTo(rectRight, rectTop)
	Path.LineTo(rectLeft, rectTop)
	cvs.DrawPath(Path, obj.fillcolor, True, 0)
	cvs.DrawPath(Path, obj.bordercolor, False, obj.strokewidth)

	If obj.shapeselector <>  obj.pfdobject Then
		PFDCommon.DrawValueInObject(cvs, obj)
	end if

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart internal storage shape.
Public Sub DrawInternalStorage(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int
	Dim offset As Float = 0.10

	'Draw the rectangle
	Dim p As B4XPath

	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.Initialize(x1, y1)
	'BotLeft
	x2 = x1
	y2 = y1 + obj.height
	p.LineTo(x2, y2)
	'BotRight
	x2 = x1 + obj.width
	y2 = y2
	p.LineTo(x2, y2)
	'TopRight
	x2 = x2
	y2 = y2 - obj.height
	p.LineTo(x2, y2)
	'TopLeft
	x2 = x1
	y2 = y1
	p.LineTo(x2, y2)
	
	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Vertical Line: TopLeft with offset to BotLeft
	x1 = obj.centerx - (obj.width/2) + (obj.width * offset)
	y1 = obj.centery - (obj.height/2)
	x2 = x1
	y2 = obj.centery + (obj.height/2)
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth/2)

	'Horizontal Line: TopLeft with offset to BotRight
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2) + (obj.width * offset)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery - (obj.height/2) + (obj.width * offset)
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth/2)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart Junction symbol.
Public Sub DrawJunction(cvs As AsyncCanvas, obj As TPFDObject)
	Dim w As Int = obj.width
	If obj.width > obj.height Then w = obj.height
	w = w - obj.strokewidth
	Dim radius As Float = w/2
	
	cvs.DrawCircle(obj.centerx, obj.centery, radius, obj.fillcolor, True, 0)
	cvs.DrawCircle(obj.centerx, obj.centery, radius, obj.bordercolor, False, obj.strokewidth)

	' Draw Line 1 (from 45° to 225°)
	DrawLineWithAngle(cvs, obj.centerX, obj.centerY, radius, 45, 225, obj.bordercolor, obj.strokewidth)

	' Draw Line 2 (from 315° to 135°)
	DrawLineWithAngle(cvs, obj.centerX, obj.centerY, radius, 315, 135, obj.bordercolor, obj.strokewidth)
	
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart loop limit shape.
Public Sub DrawLoopLimit(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath
	Dim offsetx As Float = 0.2
	Dim offsety As Float = 0.3

	'BotLeft
	p.Initialize(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))
	'BotLeft to BotRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	'BotRight to MidRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2) + (obj.height * offsety))
	'MidRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2) - (obj.width * offsetx), obj.centery - (obj.height/2))
	'TopRight to TopLeft
	p.LineTo(obj.centerx - (obj.width/2) + (obj.width * offsetx), obj.centery - (obj.height/2))
	'TopLeft to MidLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2) + (obj.height * offsety))
	'MidLeft to BotLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))

	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart manual input.
Public Sub DrawManualInput(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopRight
	p.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to Topleft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/3))
	'TopLeft to BottomLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))
	'BottomLeft to BottomRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	'BottomRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart manual loop shape.
Public Sub DrawManualLoop(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopRight
	p.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to Topleft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'TopLeft to BottomLeft
	p.LineTo(obj.centerx - (obj.width/3), obj.centery + (obj.height/2))
	'BottomLeft to BottomRight
	p.LineTo(obj.centerx + (obj.width/3), obj.centery + (obj.height/2))
	'BottomRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart merge shape.
Public Sub DrawMerge(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath
	
	'TopLeft
	p.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'BotMid
	p.LineTo(obj.centerx, obj.centery + (obj.height/2))
	'TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart offpage shape.
Public Sub DrawOffPage(cvs As AsyncCanvas, obj As TPFDObject)
	Dim offset As Float = 0.4
	Dim p As B4XPath

	'Create offpage shape with arrow down
	'BotMid
	p.Initialize(obj.centerx, obj.centery + (obj.height/2))
	'BotMid to RightMid
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2) + (obj.height * offset))
	'RightMid to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to TopLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'TopLeft to MidLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2) + (obj.height * offset))
	'MidLeft to BotMid
	p.LineTo(obj.centerx, obj.centery + (obj.height/2))
						
	'Select the direction and draw the inner and outer path
	Select obj.direction
		Case PFDCommon.DIRECTION_NONE
			cvs.DrawPath(p, obj.fillcolor, True, 0)
			cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		Case PFDCommon.DIRECTION_DOWN
			cvs.DrawPath(p, obj.fillcolor, True, 0)
			cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		Case PFDCommon.DIRECTION_UP
			cvs.DrawPathRotated(p, obj.fillcolor, True, obj.strokewidth, 180, obj.centerx, obj.centery)
			cvs.DrawPathRotated(p, obj.bordercolor, False, obj.strokewidth, 180, obj.centerx, obj.centery)
		Case PFDCommon.DIRECTION_RIGHT
			cvs.DrawPathRotated(p, obj.fillcolor, True, obj.strokewidth, 270, obj.centerx, obj.centery)
			cvs.DrawPathRotated(p, obj.bordercolor, False, obj.strokewidth, 270, obj.centerx, obj.centery)
		Case PFDCommon.DIRECTION_LEFT
			cvs.DrawPathRotated(p, obj.fillcolor, True, obj.strokewidth, 90, obj.centerx, obj.centery)
			cvs.DrawPathRotated(p, obj.bordercolor, False, obj.strokewidth, 90, obj.centerx, obj.centery)
	End Select
	
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart operation (manual operation) shape.
Public Sub DrawOperation(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopRight
	p.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to Topleft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'TopLeft to BottomLeft
	p.LineTo(obj.centerx - (obj.width/3), obj.centery + (obj.height/2))
	'BottomLeft to BottomRight
	p.LineTo(obj.centerx + (obj.width/3), obj.centery + (obj.height/2))
	'BottomRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart OR shape.
Public Sub DrawOR(cvs As AsyncCanvas, obj As TPFDObject)
	Dim w As Int = obj.width
	If obj.width > obj.height Then w = obj.height
	w = w - obj.strokewidth
	Dim radius As Float = w/2
	
	'Draw the inner and outer circle
	cvs.DrawCircle(obj.centerx, obj.centery, radius, obj.fillcolor, True, 0)
	cvs.DrawCircle(obj.centerx, obj.centery, radius, obj.bordercolor, False, obj.strokewidth)

	'Draw Line Vertical (from 0° to 180°)
	DrawLineWithAngle(cvs, obj.centerX, obj.centerY, radius, 0, 180, obj.bordercolor, obj.strokewidth/2)

	'Draw Line Horizontal (from 90° to 270°)
	DrawLineWithAngle(cvs, obj.centerX, obj.centerY, radius, 90, 270, obj.bordercolor, obj.strokewidth/2)
	
	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw shape flowchart preparation with border- and fillcolor plus text.
'Object Properties:
'value,centerx,centerx,bordercolor,fillcolor,strokewidth,textcolor,textalignment
Public Sub DrawPreparation(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,r As Int
	
	r = obj.height/2

	Dim p As B4XPath
	'LeftMid
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery
	p.Initialize(x1,y1)
	'LeftMid to LeftTop
	x1 = obj.centerx - (obj.width/2) + r
	y1 = obj.centery - (obj.height/2)
	p.LineTo(x1,y1)
	'LeftTop to RightTop
	x1 = obj.centerx + (obj.width/2) - r
	y1 = obj.centery - (obj.height/2)
	p.LineTo(x1,y1)
	'RightTop to RightMid
	x1 = obj.centerx + (obj.width/2)
	y1 = obj.centery
	p.LineTo(x1,y1)
	'RightMid to RightBot
	x1 = obj.centerx + (obj.width/2) - r
	y1 = obj.centery + (obj.height/2)
	p.LineTo(x1,y1)
	'RightBot to LeftBot
	x1 = obj.centerx - (obj.width/2) + r
	y1 = obj.centery + (obj.height/2)
	p.LineTo(x1,y1)
	'LeftBot to LeftMid
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery
	p.LineTo(x1,y1)
	cvs.DrawPath(p,obj.fillcolor,True,0)
	cvs.DrawPath(p,obj.bordercolor,False,obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw shape flowchart process with border- and fillcolor plus text.
'Object Properties:
'value,centerx,centerx,bordercolor,fillcolor,strokewidth,textcolor,textalignment
Public Sub DrawProcess(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopRight
	p.Initialize(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'TopRight to Topleft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery - (obj.height/2))
	'TopLeft to BottomLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery + (obj.height/2))
	'BottomLeft to BottomRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
	'BottomRight to TopRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart sort shape.
Public Sub DrawSort(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath

	'TopMid
	p.Initialize(obj.centerx, obj.centery - (obj.height/2))
	'MidLeft
	p.LineTo(obj.centerx - (obj.width/2), obj.centery)
	'BotMid
	p.LineTo(obj.centerx, obj.centery + (obj.height/2))
	'MidRight
	p.LineTo(obj.centerx + (obj.width/2), obj.centery)
	'TopMid
	p.LineTo(obj.centerx, obj.centery - (obj.height/2))
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
	
	'Draw the circle midline
	Dim x1,y1,x2,y2 As Int
	x1 = obj.centerx - (obj.width/2) + obj.strokewidth/2
	y1 = obj.centery
	x2 = obj.centerx + (obj.width/2) - obj.strokewidth/2
	y2 = obj.centery
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor, obj.strokewidth/2)

	PFDCommon.DrawValueInObject(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart subroutine shape.
Public Sub DrawSubRoutine(cvs As AsyncCanvas, obj As TPFDObject)
	Dim p As B4XPath
	Dim x1,y1 As Int

	'Draw rect
	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.Initialize(x1, y1)
	'TopLeft to BottomLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery + (obj.height/2)
	p.LineTo(x1, y1)
	'BottomLeft to BottomRight
	x1 = obj.centerx + (obj.width/2)
	y1 = obj.centery + (obj.height/2)
	p.LineTo(x1, y1)
	'BottomRight to TopRight
	x1 = obj.centerx + (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.LineTo(x1, y1)
	'TopRight to TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.LineTo(x1, y1)
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Draw the two inner lines left & right
	cvs.DrawLine(obj.centerx - (obj.width/2) + (obj.width/8), _
				 obj.centery - (obj.height/2) + obj.strokewidth/2, _ 
				 obj.centerx - (obj.width/2) + (obj.width/8), _
				 obj.centery + (obj.height/2) - obj.strokewidth/2, _ 
				 obj.bordercolor, obj.strokewidth/2)
	cvs.DrawLine(obj.centerx + (obj.width/2) - (obj.width/8), _
				 obj.centery - (obj.height/2) + obj.strokewidth/2, _ 
				 obj.centerx + (obj.width/2) - (obj.width/8), _
				 obj.centery + (obj.height/2) - obj.strokewidth/2, _ 
				 obj.bordercolor, obj.strokewidth/2)

	PFDCommon.DrawValueInObject(cvs, obj)
		
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw flowchart Terminator
Public Sub DrawTerminator(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int
	Dim p As B4XPath
	Dim rect As B4XRect

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rect.Initialize(x1,y1,x2,y2)
	p.InitializeRoundedRect(rect, obj.height/2)
	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Show the value in the object middle
	If obj.value.Length > 0 Then
		Dim rectvalue As B4XRect = cvs.MeasureText(obj.value, fnt)
		Dim oxvalue As Int = obj.centerx
		Dim oyvalue As Int = obj.centery + (rectvalue.height/2)
		Dim strvalue As String = obj.value
		If rectvalue.Width > obj.width Then
			strvalue = strvalue.Replace(" ", CRLF)
			oxvalue = obj.centerx
			oyvalue = obj.centery - (rectvalue.height/2)
		End If
		cvs.DrawText(strvalue, oxvalue, oyvalue, fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Helper
Private Sub DrawLineWithAngle(Cvs As AsyncCanvas, centerX As Float, centerY As Float, radius As Float, angle1 As Float, angle2 As Float, color As Int, strokewidth As Int)
	' Convert angles to radians
	Dim radian1 As Float = angle1 * cPI / 180
	Dim radian2 As Float = angle2 * cPI / 180

	' Calculate the start and end points based on angles
	Dim startX As Float = centerX + radius * Cos(radian1)
	Dim startY As Float = centerY + radius * Sin(radian1)
	Dim endX As Float = centerX + radius * Cos(radian2)
	Dim endY As Float = centerY + radius * Sin(radian2)

	' Draw the line between the calculated points
	Cvs.DrawLine(startX, startY, endX, endY, color, strokewidth)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region
