﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=10
@EndOfDesignText@
#Region Code Module Information
'Static Code Module: PFDShapesPID
'File: PFDShapesPID.bas
'Static code module for the PFD Process Instrumentation Diagram object & shapes.
'The pid objects text is set either from the tag or value depending object type.
'There are indicators with fixed state and actuators which state can be changed.
'ISA-101
'Various shapes are defined aligned to the ISA-101 standard. The settings are hardcoded in the drawshape method.
'
'PFDCreator Copyright see B4XMainPage.
#End Region

Sub Process_Globals
	'Core
	Private xui As XUI

	'PFD objects specific to this PFD flow type (MUST BE UNIQUE ACROSS THE APPLICATION).
	'Each of the PFD objects (UPPERCASE) must match with external shapes file PFDPID.shp!
	'Common PFD objects are defined in the module PFDShapesBasic.
	'Make Public to access from PFDOperator application..
	Public PID_SHAPE_BOILER As String					= "BOILER"
	Public PID_SHAPE_COLUMN As String					= "COLUMN"
	Public PID_SHAPE_COOLER As String					= "COOLER"
	Public PID_SHAPE_CONTROLLER As String				= "CONTROLLER"
	Public PID_SHAPE_FILTER As String					= "FILTER"
	Public PID_SHAPE_HEATER As String					= "HEATER"
	Public PID_SHAPE_HEAT_EXCHANGER As String			= "HEATEXCHANGER"
	Public PID_SHAPE_IMAGEPID As String					= "IMAGEPID"
	Public PID_SHAPE_INDICATOR As String				= "INDICATOR"
	Public PID_SHAPE_INSTRUMENT As String				= "INSTRUMENT"
	Public PID_SHAPE_LEVEL_INDICATOR As String			= "LEVELINDICATOR"
	Public PID_SHAPE_MIXER As String					= "MIXER"
	Public PID_SHAPE_TANK As String						= "TANK"
	Public PID_SHAPE_PIPE As String						= "PIPE"
	Public PID_SHAPE_PUMP As String						= "PUMP"
	Public PID_SHAPE_SWITCH As String					= "SWITCH"
	Public PID_SHAPE_TREND_INDICATOR As String			= "TRENDINDICATOR"
	Public PID_SHAPE_VALVE As String					= "VALVE"
	Public PID_SHAPE_VALVE3WAY As String				= "VALVE3WAY"
	Public PID_SHAPE_VESSEL As String					= "VESSEL"
	
	'ISA-101 Shapes
	Public PID_SHAPE_BINARY_STATE_INDICATOR As String	= "BINARYSTATEINDICATOR"
	Public PID_SHAPE_DATA_CONNECTOR_LINE As String		= "DATACONNECTORLINE"
	Public PID_SHAPE_NUMERIC_DATA As String				= "NUMERICDATA"
	Public PID_SHAPE_PRIMARY_PROCESS_LINE As String		= "PRIMARYPROCESSLINE"
	Public PID_SHAPE_SECONDARY_PROCESS_LINE As String	= "SECONDARYPROCESSLINE"
End Sub

#Region DrawObject
'Draw shape(s) selected from pfdobjects list.
'The shapes are defined as process_globals.
Public Sub DrawShapesFromList(cvs As AsyncCanvas, pfdobjects As List)
	'Loop over the pfd objects and draw the object shape
	For Each pfdobject As TPFDObject In pfdobjects
		DrawShape(cvs, pfdobject)
	Next
End Sub

'Draw a shape by shape type as defined in process_globals.
Public Sub DrawShape(cvs As AsyncCanvas, obj As TPFDObject)
	Select obj.pfdobject
			'PFD Specific Types
		Case PID_SHAPE_BOILER:
			DrawBoiler(cvs, obj)
		Case PID_SHAPE_COLUMN:
			DrawColumn(cvs, obj)
		Case PID_SHAPE_COOLER
			DrawCooler(cvs, obj)
		Case PID_SHAPE_CONTROLLER:
			DrawController(cvs, obj)
		Case PID_SHAPE_FILTER
			DrawFilter(cvs, obj)
		Case PID_SHAPE_HEATER
			DrawHeater(cvs, obj)
		Case PID_SHAPE_HEAT_EXCHANGER
			DrawHeatExchanger(cvs, obj)
		Case PID_SHAPE_IMAGEPID:
			DrawImagePID(cvs, obj)
		Case PID_SHAPE_INDICATOR:
			DrawIndicator(cvs, obj)
		Case PID_SHAPE_INSTRUMENT:
			DrawInstrument(cvs, obj)
		Case PID_SHAPE_LEVEL_INDICATOR:
			DrawLevelIndicator(cvs, obj)
		Case PID_SHAPE_MIXER:
			DrawMixer(cvs, obj)
		Case PID_SHAPE_TANK:
			DrawTank(cvs, obj)
		Case PID_SHAPE_PUMP:
			DrawPump(cvs, obj)
		Case PID_SHAPE_SWITCH:
			DrawSwitch(cvs, obj)
		Case PID_SHAPE_TREND_INDICATOR:
			DrawTrendIndicator(cvs, obj)
		Case PID_SHAPE_VALVE:
			DrawValve(cvs, obj)
		Case PID_SHAPE_VALVE3WAY:
			DrawValve3Way(cvs, obj)
		Case PID_SHAPE_VESSEL:
			DrawVessel(cvs, obj)
		
		'ISA101
		Case PID_SHAPE_BINARY_STATE_INDICATOR:
			DrawBinaryStateIndicator(cvs, obj)
		Case PID_SHAPE_DATA_CONNECTOR_LINE:
			DrawDataConnectorLine(cvs, obj)
		Case PID_SHAPE_NUMERIC_DATA:
			DrawNumericData(cvs, obj)
		Case PID_SHAPE_PRIMARY_PROCESS_LINE:
			DrawPrimaryProcessLine(cvs, obj)
		Case PID_SHAPE_SECONDARY_PROCESS_LINE:
			DrawSecondaryProcessLine(cvs, obj)

	End Select
End Sub
#End Region

'*******************************************************************************
'Define object drawings in the next regions.
'Each object drawing method has 2 arguments AsyncCanvas and TPFDObject.
'Ensure all object properties are set.
'Use PFDCommon methods, like SetFont, DrawConnector, 
'DrawTag(cvs,obj), DrawTagValue(cvs,obj)

'Commit the drawing is done in method ReDraw (cvs.Invalidate).
'
'Example simple drawtext:
'Public Sub DrawOBJECT(cvs As AsyncCanvas, obj As TPFDObject)
'	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
'	cvs.DrawText(obj.tag, obj.centerx, obj.centery, fnt, obj.textcolor, obj.textalignment)
'End Sub
'HINT
'It is helpful to draw a rectangle to check the object drawing in range.
'Use this code at the end of the DrawOBJECT method.
'Dim rect As B4XRect
'rect.Initialize(obj.centerx - (obj.width/2), obj.centery - (obj.height/2), _
'				 obj.centerx + (obj.width/2), obj.centery + (obj.height/2))
'cvs.DrawRect(rect, xui.Color_Blue, False, 0.5)
'*******************************************************************************

#Region Boiler
'Draw a boiler.
'PID Abbreviation: B
Public Sub DrawBoiler(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x0, y0, x1,y1,x2,y2,x3,y3 As Int

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	'Dim fntrect As B4XRect = cvs.MeasureText(obj.tag, fnt)

	Dim segment As Int = obj.height / 10

	'Left starting at the bottom
	x0 = obj.centerx - (obj.width/2) + obj.strokewidth
	y0 = obj.centery + (obj.height/2)

	x1 = x0
	y1 = y0 - (segment * 7)
	cvs.DrawLine(x0, y0, x1, y1, obj.bordercolor, obj.strokewidth)

	x2 = x1 + (obj.width/4)
	y2 = y1 - (segment * 1)
	cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor, obj.strokewidth)

	x3 = x2
	y3 = y2 - (segment * 2)
	cvs.DrawLine(x2, y2, x3, y3, obj.bordercolor, obj.strokewidth)

	'Right starting at the bottom
	x0 = obj.centerx + (obj.width/2)  - obj.strokewidth
	y0 = obj.centery + (obj.height/2)

	x1 = x0
	y1 = y0 - (segment * 7)
	cvs.DrawLine(x0, y0, x1, y1, obj.bordercolor, obj.strokewidth)

	x2 = x1 - (obj.width/4)
	y2 = y1 - (segment * 1)
	cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor, obj.strokewidth)

	x3 = x2
	y3 = y2 - (segment * 2)
	cvs.DrawLine(x2, y2, x3, y3, obj.bordercolor, obj.strokewidth)

	'Bottom line
	x1 = obj.centerx - (obj.width/2) + obj.strokewidth
	y1 = obj.centery + (obj.height/2) - (obj.strokewidth/2)
	x2 = obj.centerx + (obj.width/2) - obj.strokewidth
	y2 = obj.centery + (obj.height/2) - (obj.strokewidth/2)
	cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor, obj.strokewidth)

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER
	cvs.DrawText(obj.tag, obj.centerx, obj.centery, fnt, obj.textcolor, obj.textalignment)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Column
'Draw a column which is a rounded rect.
'PID Abbreviation: C
Public Sub DrawColumn(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int
	Dim p As B4XPath
	Dim rect As B4XRect

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Check the direction
	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_VERTICAL
		If obj.width > obj.height Then
			Dim w As Int = obj.width
			Dim h As Int = obj.height
			obj.height = w
			obj.width = h
		End If
	End If
	
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rect.Initialize(x1,y1,x2,y2)

	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		p.InitializeRoundedRect(rect, obj.height/2)
	End If
	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		p.InitializeRoundedRect(rect, obj.width/2)
	End If

	'Draw the inner path
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer path
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Show the value in the object middle
	If obj.value.Length > 0 Then
		Dim rectvalue As B4XRect = cvs.MeasureText(obj.value, fnt)
		Dim oxvalue As Int = obj.centerx
		Dim oyvalue As Int = obj.centery + (rectvalue.height/2)
		Dim strvalue As String = obj.value
		If rectvalue.Width > obj.width Then
			strvalue = strvalue.Replace(" ", CRLF)
			oxvalue = obj.centerx
			oyvalue = obj.centery - (rectvalue.height/2)
		End If
		cvs.DrawText(strvalue, oxvalue, oyvalue, fnt, obj.textcolor, obj.textalignment)
	End If

	'Draw the connection lines
	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		x1 = obj.centerx - (obj.width/2) + (obj.strokewidth/2)
		y1 = obj.centery - (obj.height/2) + (obj.height/4)
		x2 = obj.centerx + (obj.width/2) - (obj.strokewidth/2)
		y2 = y1
		cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth)
		
		x1 = obj.centerx - (obj.width/2) + (obj.strokewidth/2)
		y1 = obj.centery + (obj.height/2) - (obj.height/4)
		x2 = obj.centerx + (obj.width/2) - (obj.strokewidth/2)
		y2 = y1
		cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth)
	End If
	
	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		x1 = obj.centerx - (obj.width/4)
		y1 = obj.centery - (obj.height/2)
		x2 = x1
		y2 = obj.centery + (obj.height/2)
		cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth)
		x1 = obj.centerx + (obj.width/4)
		y1 = obj.centery - (obj.height/2)
		x2 = x1
		y2 = obj.centery + (obj.height/2)
		cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth)
	End If

	PFDCommon.DrawTag2(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Controller
'Draw an controller with 3 text lines top,mid,bot with seperator.
'top=tag, separator, mid=setpoint sp value, bot=process pv value.
'The value is a cvs string with value sptext,spvalue,pvtext,pvvalue:
'sp:,12,pv,11
'PID Abbreviation: FR, FI etc.
Public Sub DrawController(cvs As AsyncCanvas, obj As TPFDObject)

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText(obj.tag, fnt)
	'Tag font height
	Dim th As Int = Round(fntrect.Height)
	'Tag font height margin calculated from the textbox height
	Dim tm As Int = (obj.height - th) / 2
	'Dim tl As Int = fntrect.Width

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Draw the rectangle
	Dim x1,y1,x2,y2 As Int
	Dim p As B4XPath

	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.Initialize(x1, y1)
	'BotLeft
	x2 = x1
	y2 = y1 + obj.height
	p.LineTo(x2, y2)
	'BotRight
	x2 = x1 + obj.width
	y2 = y2
	p.LineTo(x2, y2)
	'TopRight
	x2 = x2
	y2 = y2 - obj.height
	p.LineTo(x2, y2)
	'TopLeft
	x2 = x1
	y2 = y1
	p.LineTo(x2, y2)
	
	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)
		
	'Draw the tag
	x1 = obj.centerx
	y1 = obj.centery - (obj.height/2) + (obj.height/3) - (tm/6)
	cvs.DrawText(obj.tag, x1, y1, fnt, obj.textcolor, obj.textalignment)
	
	'Draw the separator line
	x1 = obj.centerx - (obj.width/2) + obj.strokewidth
	y1 = obj.centery - (obj.height/2) + (obj.height/3)
	x2 = obj.centerx + (obj.width/2) - obj.strokewidth
	y2 = y1
	cvs.DrawLine(x1, y1, x2, y1, obj.bordercolor, obj.strokewidth/2)

	'The value is a cvs string with value sptext,spvalue,pvtext,pvvalue.
	Dim sptext As String = ""	'set point
	Dim spvalue As String = ""	'set point
	Dim pvtext As String = ""	'process value
	Dim pvvalue As String = ""	'set point
	
	'Check if there is a value else set default
	If obj.value == "" Then obj.value = "sp:, ,pv, "
	
	Dim values() As String = Regex.Split(",", obj.value)
	If values.Length <> 4 Then
		Log($"[ERROR][DrawController] Value definition incorrect. Set sptext,spvalue,pvtext,pvvalue"$)
		x1 = obj.centerx
		y1 = obj.centery - (obj.height/2) + ( (obj.height/3) * 2) - (tm/6)
		cvs.DrawText($"ERROR"$, x1, y1, fnt, obj.textcolor, obj.textalignment)
		Return
	End If
	'Assign the csv values
	sptext = values(0)
	spvalue = values(1)
	pvtext = values(2)
	pvvalue = values(3)

	'Draw the setpoint sp as text+value+unit
	x1 = obj.centerx
	y1 = obj.centery - (obj.height/2) + ( (obj.height/3) * 2) - (tm/6)
	cvs.DrawText($"${sptext}${spvalue}${obj.unit}"$, x1, y1, fnt, obj.textcolor, obj.textalignment)
	'Draw the sep
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2) + ( (obj.height/3) * 2)
	x2 = obj.centerx + (obj.width/2)
	y2 = y1
	cvs.DrawLine(x1, y1, x2, y1, obj.bordercolor, obj.strokewidth/2)

	'Draw the process value pv as text+value+unit
	x1 = obj.centerx
	y1 = obj.centery - (obj.height/2) + ( (obj.height/3) * 3) - (tm/6)
	cvs.DrawText($"${pvtext}${pvvalue}${obj.unit}"$, x1, y1, fnt, obj.textcolor, obj.textalignment)

	'Draw the object connector (if object direction is set)
	PFDShapesBasic.DrawConnector(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Cooler
'Draw a cooler which is a heat exchanger with cross arrow showing top right (up).
'Object width and height must be the same.
'PID Abbreviation: E
Public Sub DrawCooler(cvs As AsyncCanvas, obj As TPFDObject)
	'Draw the heatexchanger
	DrawHeatExchanger(cvs, obj)

	'Rotated arrow object with line
	Dim objal As TPFDObject
	objal.Initialize
	objal.centerx = obj.centerx
	objal.centery = obj.centery
	objal.width = obj.width
	objal.height = objal.width
	objal.value = -45	'=right up, 45=right down
	objal.strokewidth = 2
	objal.fillcolor = obj.bordercolor
	objal.direction = PFDCommon.DIRECTION_UP
	PFDShapesBasic.DrawArrowSolidRotated(cvs, objal)

	'Draw the tag
	PFDCommon.DrawTag(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Filter
'Draw a Filter which is a rect with border and dotted line from bottomleft to topright.
'Object properties:
'tag,centerx,centery,width,height,strokewidth,bordercolor,fillcolor,textcolor,textalignment,seperator,direction
'PID Abbreviation: F
Public Sub DrawFilter(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int

	'Draw the rectangle
	Dim x1,y1,x2,y2 As Int
	Dim p As B4XPath

	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.Initialize(x1, y1)
	'BotLeft
	x2 = x1
	y2 = y1 + obj.height
	p.LineTo(x2, y2)
	'BotRight
	x2 = x1 + obj.width
	y2 = y2
	p.LineTo(x2, y2)
	'TopRight
	x2 = x2
	y2 = y2 - obj.height
	p.LineTo(x2, y2)
	'TopLeft
	x2 = x1
	y2 = y1
	p.LineTo(x2, y2)
	
	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Draw the dotted line from  bottomleft to topright
	'bottomleft
	x1 = obj.centerx - (obj.width / 2) + (obj.strokewidth)
	y1 = obj.centery + (obj.height / 2) - (obj.strokewidth)
	'topright
	x2 = obj.centerx + (obj.width / 2) - (obj.strokewidth)
	y2 = obj.centery - (obj.height / 2) + (obj.strokewidth)

	' Calculate the length of the diagonal
	Dim dx As Float = x2 - x1
	Dim dy As Float = y2 - y1
	Dim distance As Float = Sqrt(dx * dx + dy * dy)

	' Set the size of the dashes and the gap between them
	Dim dashLength As Float = 2dip
	Dim gap As Float = 2dip

	' Calculate the unit vector for direction along the diagonal
	Dim unitX As Float = dx / distance
	Dim unitY As Float = dy / distance

	' Loop through the diagonal and draw small line segments (dashes)
	Dim numberOfDashes As Int = distance / (dashLength + gap)
	
	For i = 0 To numberOfDashes - 1
		' Calculate the starting position of each dash
		Dim fractionStart As Float = i / numberOfDashes
		Dim xx1 As Float = x1 + fractionStart * dx	'left
		Dim yy1 As Float = y1 + fractionStart * dy	'top

		' Calculate the ending position of each dash
		Dim xx2 As Float = xx1 + unitX * dashLength
		Dim yy2 As Float = yy1 + unitY * dashLength

		' Draw the dash (small line)
		cvs.DrawLine(xx1, yy1, xx2, yy2, obj.bordercolor, obj.strokewidth)
	Next

	'Draw the tag
	PFDCommon.DrawTag(cvs, obj)
	
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Heater
'Draw a heater which is a heat exchanger with cross arrow showing bottom left (down).
'Object width and height must be the same.
'PID Abbreviation: E
Public Sub DrawHeater(cvs As AsyncCanvas, obj As TPFDObject)
	'Draw the heatexchanger
	DrawHeatExchanger(cvs, obj)

	'Rotated arrow object with line
	Dim objal As TPFDObject
	objal.Initialize
	objal.centerx = obj.centerx
	objal.centery = obj.centery
	objal.width = obj.width
	objal.height = objal.width
	objal.value = 135 '45=right down, -45=right up
	objal.strokewidth = 2
	objal.fillcolor = obj.bordercolor
	objal.direction = PFDCommon.DIRECTION_DOWN
	PFDShapesBasic.DrawArrowSolidRotated(cvs, objal)

	'Draw the tag
	PFDCommon.DrawTag(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region HeatExchanger
'Draw an heat exchanger.
'Object width and height must be the same.
'PID Abbreviation: E
Public Sub DrawHeatExchanger(cvs As AsyncCanvas, obj As TPFDObject)
	'Inner pipe pos from left to right or up to down
	Dim x1,y1,x2,y2,x3,y3,x4,y4,x5,y5,x6,y6 As Int
	Dim w, h As Int
	
	'Check & adjust the width which is used for the circle.
	w = obj.width - obj.strokewidth
	h = obj.height - obj.strokewidth

	'Draw the inner shape
	cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.bordercolor, False, obj.strokewidth)

	'Check direction
	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If

	'Draw the exchanger inner pipe with default strokewidth 2
	Dim sw As Int = 2
	Select obj.direction
		Case PFDCommon.DIRECTION_HORIZONTAL
			x1 = obj.centerx - (w/2)
			y1 = obj.centery
			x2 = x1 + (w * 0.1)
			y2 = y1
			cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,sw)
			
			x3 = x1 + (w * 0.25)
			y3 = y1 - (h * 0.25)
			cvs.DrawLine(x2,y2,x3,y3,obj.bordercolor,sw)
			
			x4 = x1 + (w * 0.75)
			y4 = y1 + (h * 0.25)
			cvs.DrawLine(x3,y3,x4,y4,obj.bordercolor,sw)
			
			x5 = x1 + (w * 0.9)
			y5 = y1
			cvs.DrawLine(x4,y4,x5,y5,obj.bordercolor,sw)
			
			x6 = x1 + (w)
			y6 = y1
			cvs.DrawLine(x5,y5,x6,y6,obj.bordercolor,sw)
			
		Case PFDCommon.DIRECTION_VERTICAL
			x1 = obj.centerx
			y1 = obj.centery - (h/2)
			x2 = x1
			y2 = y1 + (h * 0.1)
			cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,sw)

			x3 = x1 + (w * 0.25)
			y3 = y1 + (h * 0.25)
			cvs.DrawLine(x2,y2,x3,y3,obj.bordercolor,sw)

			x4 = x1 - (w * 0.25)
			y4 = y1 + (h * 0.75)
			cvs.DrawLine(x3,y3,x4,y4,obj.bordercolor,sw)

			x5 = x1 
			y5 = y1 + (h * 0.9)
			cvs.DrawLine(x4,y4,x5,y5,obj.bordercolor,sw)

			x6 = x1
			y6 = y1 + (h)
			cvs.DrawLine(x5,y5,x6,y6,obj.bordercolor,sw)
		
	End Select

	'Draw the tag
	PFDCommon.DrawTag(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Image
'Draw an image with tag in the image middle.
Public Sub DrawImagePID(cvs As AsyncCanvas, obj As TPFDObject)
	PFDShapesBasic.DrawImage(cvs, obj)
	PFDCommon.DrawTag2(cvs, obj)
End Sub
#End Region

#Region Indicator
'Draw an indicator which is a circle with optional seperator.
'The indicator has 2 text lines:
'Top line: obj.tag, Bottom line: obj.value + obj.unit
'PID Abbreviation: I
Public Sub DrawIndicator(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x,y As Int
	Dim offset As Int = 2

	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntsmall As B4XFont = PFDCommon.SetFontSmall(obj)
	Dim fntrect As B4XRect = cvs.MeasureText(obj.tag, fnt)

	Dim w As Int = obj.width - obj.strokewidth

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Draw the inner shape
	cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.fillcolor, True, 0)
	'Draw the outer shape
	If obj.strokewidth > 0 Then cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.bordercolor, False, obj.strokewidth)

	'Draw the tag
	cvs.DrawText(obj.tag, obj.centerx, obj.centery - (fntrect.Height/4), fnt, obj.textcolor, obj.textalignment)

	'Draw the separator
	If obj.separator Then
		cvs.DrawLine(obj.centerx - w/2 + (obj.strokewidth/2), obj.centery, _ 
					 obj.centerx + w/2 - (obj.strokewidth/2), obj.centery, _ 
					 obj.bordercolor, obj.strokewidth/2)
	End If

	'Draw the value and unit
	'PFDShapesBasic.DrawValueUnit2(cvs, obj)
	x = obj.centerx
	y = obj.centery + fntrect.Height + (obj.strokewidth) + offset
	cvs.DrawText($"${obj.value}"$, x, y, fnt, obj.textcolor, obj.textalignment)
	x = x
	y = y + fntrect.height + offset
	cvs.DrawText($"${obj.unit}"$, x, y, fntsmall, obj.textcolor, obj.textalignment)

	'Draw the object connector (if object direction TOP, BOTTOM, LEFT, RIGHTis set)
	PFDShapesBasic.DrawConnector(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw an indicator which is a circle with optional separator.
'The indicator text can have 1 or 2 lines as defined in the object value.
'1 line: obj.value = "FI", 2 lines: obj.value = $"FI${CRLF}123"$.
'PID Abbreviation: I
Public Sub DrawIndicator2(cvs As AsyncCanvas, obj As TPFDObject)
	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText(obj.tag, fnt)

	Dim w As Int = obj.width - obj.strokewidth

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Draw the inner shape
	cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawCircle(obj.centerx, obj.centery, w/2, obj.bordercolor, False, obj.strokewidth)

	'Draw the separator
	If obj.separator Then
		cvs.DrawLine(obj.centerx - w/2 + (obj.strokewidth/2), obj.centery, _ 
					 obj.centerx + w/2 - (obj.strokewidth/2), obj.centery, _ 
					 obj.bordercolor, obj.strokewidth)
	End If

	'Draw the text 1 or 2 lines depending CRLF
	Dim s() As String = Regex.Split(CRLF, obj.tag)
	If s.Length == 1 Then
		cvs.DrawText(s(0), obj.centerx, obj.centery, fnt, obj.textcolor, obj.textalignment)
	End If

	If s.Length == 2 Then
		Dim y As Int
		y = obj.centery - (w/4) + (fntrect.Height/4)
		cvs.DrawText(s(0), obj.centerx, y, fnt, obj.textcolor, obj.textalignment)
		y = obj.centery + (w/4) + (fntrect.Height/8)
		cvs.DrawText(s(1), obj.centerx, y, fnt, obj.textcolor, obj.textalignment)
	End If

	'Draw the object connector (if object direction is set)
	PFDShapesBasic.DrawConnector(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Instrument
'Draw an instrument with 2 text lines with seperator
'top=tag, bottom=value + unit.
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: I
Public Sub DrawInstrument(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText(obj.tag, fnt)
	'Tag font height
	Dim th As Int = Round(fntrect.Height)
	'Tag font height margin calculated from the textbox height
	Dim tm As Int = (obj.height - th) / 2

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Draw the rectangle
	Dim p As B4XPath

	'TopLeft
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	p.Initialize(x1, y1)
	'BotLeft
	x2 = x1
	y2 = y1 + obj.height
	p.LineTo(x2, y2)
	'BotRight
	x2 = x1 + obj.width
	y2 = y2
	p.LineTo(x2, y2)
	'TopRight
	x2 = x2
	y2 = y2 - obj.height
	p.LineTo(x2, y2)
	'TopLeft
	x2 = x1
	y2 = y1
	p.LineTo(x2, y2)
	
	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Draw the tag text	
	cvs.DrawText(obj.tag, x1 + (obj.width/2), y1 + tm, fnt, obj.textcolor, obj.textalignment)

	'Draw seperator line in the middle with strokewidth 1
	x1 = obj.centerx - (obj.width/2) '+ obj.strokewidth
	y1 = obj.centery
	x2 = obj.centerx + (obj.width/2) - obj.strokewidth
	y2 = y1
	cvs.DrawLine(x1, y1, x2, y2, obj.bordercolor, 1)
	
	'Draw the value and unit
	cvs.DrawText($"${obj.value}${obj.unit}"$, x1 + (obj.width/2), obj.centery + tm, fnt, obj.textcolor, obj.textalignment)
	'cvs.DrawText($"${obj.value}${obj.unit}"$, x1 + (obj.width/2), y1 + (obj.height/2) + tm, fnt, obj.textcolor, obj.textalignment)

	'Draw the object connector (if object direction is set)
	PFDShapesBasic.DrawConnector(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region LevelIndicator
'Draw a LevelIndicator which is a rect with border (bordercolor) and filled (fillcolor) area 0 - 100%,
'The background color of the bar is transparent.
'If the value is not set (=empty), then value 0 is set to ensure the bar can be drawn (even if 0).
'Object properties:
'centerx,centery,width,height,value,strokewidth,bordercolor,fillcolor,textcolor,textalignment,seperator,direction
'PID Abbreviation: LI
Public Sub DrawLevelIndicator(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int

	'Check the value, if not set (empty) then set to 0 to ensure the bar height can be calculated
	If obj.value.Length == 0 Then obj.value = 0
	If Not(Helper.ValueIsNumber(obj.value)) Then obj.value = 0

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER

	'Check direction hor or vert
	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_VERTICAL
	End If

	If obj.direction == PFDCommon.DIRECTION_VERTICAL Or obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		
		'Create the level rect
		x1 = obj.centerx - (obj.width / 2)
		y1 = obj.centery - (obj.height / 2)
		x2 = x1 + obj.width
		y2 = y1 + obj.height

		Dim rect As B4XRect
		rect.Initialize(x1,y1,x2,y2)
		'Draw the inner shape
		cvs.DrawRoundRect(rect,xui.Color_White, True, obj.strokewidth, obj.cornerradius)
		'Draw the outer shape
		cvs.DrawRoundRect(rect,obj.bordercolor, False, obj.strokewidth, obj.cornerradius)
				
		'Create a filled rect showing the level bottom up
		If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
			x1 = obj.centerx - (obj.width / 2) + (obj.strokewidth)
			y1 = obj.centery + (obj.height/2) - (obj.height * (obj.value * 0.01)) + (obj.strokewidth)
			x2 = obj.centerx + (obj.width / 2) - (obj.strokewidth)
			y2 = obj.centery + (obj.height/2) - (obj.strokewidth)
		End If

		'Create a filled rect showing the level left to right
		If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
			Dim w As Int = obj.width - (obj.strokewidth)
			x1 = obj.centerx - (obj.width/2) + (obj.strokewidth/2)
			y1 = obj.centery - (obj.height/2) + (obj.strokewidth/2)
			x2 = obj.centerx - (obj.width/2) + (w * (obj.value * 0.01)) + (obj.strokewidth/2)
			y2 = obj.centery + (obj.height/2) - (obj.strokewidth/2)
		End If
		Dim rectlevel As B4XRect
		rectlevel.Initialize(x1,y1,x2,y2)
		'Draw the level rect
		cvs.DrawRoundRect(rectlevel,obj.fillcolor, True, obj.strokewidth, obj.cornerradius)

		'Draw the tag and level below the bar
		PFDCommon.DrawTagValue(cvs, obj)

		'Commit the drawing
		'cvs.Invalidate
	End If

End Sub

Public Sub DrawLevelIndicator2(cvs As AsyncCanvas, obj As TPFDObject)
	Dim rect As B4XRect
	Dim x1,y1,x2,y2 As Int

	'Check direction hor or vert
	If obj.direction == PFDCommon.DIRECTION_VERTICAL Or obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		
		'Textbox rect topleft, bottomright
		x1 = obj.centerx - (obj.width / 2)
		y1 = obj.centery - (obj.height / 2)
		x2 = x1 + obj.width
		y2 = y1 + obj.height

		'Create the level indicator rect
		rect.Initialize(x1, y1, x2, y2)

		'Clear the level indicator rect area first
		cvs.ClearRect(rect)

		'Draw the inner shape = NOT USED
		'cvs.DrawRect(rect, obj.fillcolor, True, 0)
		'Draw the outer shape
		cvs.DrawRect(rect, obj.bordercolor, False, obj.strokewidth)

		'Check the value, if not set (empty) then set to 0 to ensure the bar height can be calculated
		If obj.value.Length == 0 Then obj.value = 0

		'Create a filled rect showing the level bottom up
		If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
			x1 = obj.centerx - (obj.width / 2) + (obj.strokewidth/2)
			y1 = obj.centery + (obj.height/2) - (obj.height * (obj.value * 0.01)) + (obj.strokewidth/2)
			x2 = obj.centerx + (obj.width / 2) - (obj.strokewidth/2)
			y2 = obj.centery + (obj.height/2) - (obj.strokewidth/2)
		End If

		'Create a filled rect showing the level left to right
		If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
			Dim w As Int = obj.width - (obj.strokewidth)
			x1 = obj.centerx - (obj.width/2) + (obj.strokewidth/2)
			y1 = obj.centery - (obj.height/2) + (obj.strokewidth/2)
			x2 = obj.centerx - (obj.width/2) + (w * (obj.value * 0.01)) + (obj.strokewidth/2)
			y2 = obj.centery + (obj.height/2) - (obj.strokewidth/2)
		End If
		
		'Create the level rect
		Dim levelrect As B4XRect
		levelrect.Initialize(x1, y1, x2, y2)
		cvs.DrawRect(levelrect, obj.fillcolor, True, obj.strokewidth)

		'Draw the level below the bar
		PFDCommon.DrawTagValue(cvs, obj)

		'Commit the drawing
		'cvs.Invalidate
	End If

End Sub
#End Region

#Region Mixer
'Draw a mixer which is a rect containing a (kind of) Y.
'Object properties:
'centerx,centery,width,height,value,strokewidth,bordercolor,fillcolor,textcolor,textalignment,seperator,direction
'PID Abbreviation: M
Public Sub DrawMixer(cvs As AsyncCanvas, obj As TPFDObject)
	Dim rect As B4XRect
	Dim x1,y1,x2,y2 As Int

	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = x1 + obj.width
	y2 = y1 + obj.height

	'Create the instrument rect
	rect.Initialize(x1, y1, x2, y2)

	'Clear the instrument rect area first
	cvs.ClearRect(rect)
	
	'Draw the inner shape
	cvs.DrawRect(rect, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawRect(rect, obj.bordercolor, False, obj.strokewidth)

	'Draw the mixer symbol = an Y
	obj.strokewidth = 2
	
	'Line Center > Down
	x1 = obj.centerx
	y1 = obj.centery
	x2 = x1
	y2 = obj.centery + (obj.height/2) - (obj.strokewidth * 3)
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,(obj.strokewidth))

	'Line Center > UpLeft
	x1 = obj.centerx
	y1 = obj.centery
	x2 = obj.centerx - (obj.width/3)
	y2 = y1 - (obj.height/4)
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,(obj.strokewidth))

	'Line Center > UpRight
	x1 = obj.centerx
	y1 = obj.centery
	x2 = obj.centerx + (obj.width/3)
	y2 = y1 - (obj.height/4)
	cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,(obj.strokewidth))

	'Draw the Tag
	PFDCommon.DrawTag(cvs, obj)

	'Draw the object connector (if object direction is set)
	PFDShapesBasic.DrawConnector(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Pipe
'Draw a pipe which is basically a thicker solid line.
'PID Abbreviation: PP
Public Sub DrawPipe(cvs As AsyncCanvas, obj As TPFDObject)
	Select obj.direction
		Case PFDCommon.DIRECTION_HORIZONTAL:
			PFDShapesBasic.DrawLineSolidHorizontal(cvs, obj)
		Case PFDCommon.DIRECTION_VERTICAL:
			PFDShapesBasic.DrawLineSolidVertical(cvs, obj)
	End Select
End Sub
#End Region

#Region Pump
'Draw a pump: circle with triangle pointing to the flow direction.
'Pump state: 0=stopped,1=running,99=error
'Object Properties:
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: P
Public Sub DrawPump(cvs As AsyncCanvas, obj As TPFDObject)
	Dim r As Int

	'Set default direction in case none
	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_RIGHT
	End If
	'Pump direction taken from the idirection UP,DOWN,LEFT,RIGHT
	Dim direction As String = obj.direction.ToUpperCase

	'Pump radius
	r = obj.width / 2

	'Draw the circle
	cvs.DrawCircle(obj.centerx, obj.centery, r, obj.bordercolor, True, obj.strokewidth)
	
	'Define the path for the pump triangle
	Dim p As B4XPath
	Dim angle As Int
	Dim x(3),y(3) As Float

	Select direction.ToUpperCase
		Case PFDCommon.DIRECTION_UP
			angle = 270 + 0
			x(0) = (CosD(angle) * r * 0.95) + obj.centerx
			y(0) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 125
			x(1) = (CosD(angle) * r * 0.95) + obj.centerx
			y(1) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 235
			x(2) = (CosD(angle) * r * 0.95) + obj.centerx
			y(2) = (SinD(angle) * r * 0.95) + obj.centery

		Case PFDCommon.DIRECTION_DOWN
			angle = 270 + 180
			x(0) = (CosD(angle) * r * 0.95) + obj.centerx
			y(0) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 305
			x(1) = (CosD(angle) * r * 0.95) + obj.centerx
			y(1) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 55
			x(2) = (CosD(angle) * r * 0.95) + obj.centerx
			y(2) = (SinD(angle) * r * 0.95) + obj.centery

		Case PFDCommon.DIRECTION_LEFT
			angle = 270 - 90
			x(0) = (CosD(angle) * r * 0.95) + obj.centerx
			y(0) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 35
			x(1) = (CosD(angle) * r * 0.95) + obj.centerx
			y(1) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 145
			x(2) = (CosD(angle) * r * 0.95) + obj.centerx
			y(2) = (SinD(angle) * r * 0.95) + obj.centery

		Case PFDCommon.DIRECTION_RIGHT
			angle = 270 + 325
			x(0) = (CosD(angle) * r * 0.95) + obj.centerx
			y(0) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 90
			x(1) = (CosD(angle) * r * 0.95) + obj.centerx
			y(1) = (SinD(angle) * r * 0.95) + obj.centery
			angle = 270 + 215
			x(2) = (CosD(angle) * r * 0.95) + obj.centerx
			y(2) = (SinD(angle) * r * 0.95) + obj.centery

	End Select

	'Create & draw the path of the triangle with statecolor and filled
	p.Initialize(x(0), y(0))
	p.LineTo(x(1), y(1))
	p.LineTo(x(2), y(2))
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	
	'Draw the pump tag & value
	PFDCommon.DrawTagValue(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Switch
'Draw a switch: outer and inner circle.
'Object Properties:
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: SW
Public Sub DrawSwitch(cvs As AsyncCanvas, obj As TPFDObject)
	Dim r As Int
	Dim w As Int

	w = obj.width - obj.strokewidth

	'Switch radius
	r = w / 2

	'Draw the inner shape = NOT USED
	cvs.DrawCircle(obj.centerx, obj.centery, r, obj.fillcolor, True, obj.strokewidth)
	'Draw the outer shape
	cvs.DrawCircle(obj.centerx, obj.centery, r, obj.bordercolor, False, obj.strokewidth)

	'Draw the tag & value
	PFDCommon.DrawTagValue(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Tank
'Draw a cone roof tank.
'Object Properties:
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: T
Public Sub DrawTank(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x,y As Int
	'Set the cone height factor from the object height
	Dim coneheightfactor As Float = 0.6
	Dim p As B4XPath
	
	'Bottomleft
	x = obj.centerx - (obj.width/2)
	y = obj.centery + (obj.height/2)
	p.Initialize(x,y)
	'Topleft
	x = obj.centerx - (obj.width/2)
	y = obj.centery - ( (obj.height / 2) * coneheightfactor)
	p.LineTo(x,y)
	'Topmiddle
	x = obj.centerx
	y = obj.centery - (obj.height/2)
	p.LineTo(x,y)
	'TopRight
	x = obj.centerx + (obj.width/2)
	y = obj.centery - ( (obj.height / 2) * coneheightfactor)
	p.LineTo(x,y)
	'Bottomright
	x = obj.centerx + (obj.width/2)
	y = obj.centery + (obj.height/2)
	p.LineTo(x,y)
	'Bottomleft
	x = obj.centerx - (obj.width/2)
	y = obj.centery + (obj.height/2)
	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Draw the tag
	PFDCommon.DrawTagValue(cvs,obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region TrendIndicator
'Draw trend indicator which is a small chart with x-axis time and y-axis value.
'The recommended obj width = 200 and height = 100
'The chart properties incl. data which is set in the obj.value as JSON.
'Important to set the strings in single quotes '' and not "":
'{
'	'title':'TEST CHART', 
'	'scaley':{'min':0,'max':100,'ticks':5},
'	'data':[{'x':1, 'y':50},{'x':2, 'y':80},{'x':3, 'y':10}]
'}
'The data x,y pairs are used to draw the point on the chart, connect lines between poinzts and set the x-axis scale.
Public Sub DrawTrendIndicator(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Float

	Dim valuejson As String = obj.value

	'Offset of the rectchart
	Dim chartoffset As Float = obj.height / 5	'20
	'Scale tick size
	Dim scaleticksize As Int = 2
	Dim scalextextrotation As Int = 0

	'Set scale font
	Dim fnt As B4XFont
	fnt = xui.CreateDefaultFont(obj.fontsize)

	'Set default value in case no value given.
	'Important to set the strings in single quotes '' and not "".
	If valuejson.length == 0 Then
		valuejson = $"{'title':'Trend','scaley':{'min':0,'max':100,'ticks':5},'data':[{'x':1, 'y':50},{'x':2, 'y':80},{'x':3, 'y':10}]}"$
		obj.value = valuejson
	End If

	'Get the properties from the value
	Dim properties As String = valuejson
	Dim parser As JSONParser
	Try
		parser.Initialize(properties)
		Dim jRoot As Map = parser.NextObject
		Dim scaley As Map = jRoot.Get("scaley")
		Dim scaleymin As Int = scaley.Get("min")
		Dim scaleymax As Int = scaley.Get("max")
		Dim scaleyticks As Int = scaley.Get("ticks")
		Dim data As List = jRoot.Get("data")
		Dim scalexticks As Int = data.size
		Dim title As String = jRoot.Get("title")		
	Catch
		Log($"[ERROR][TrendIndicator] Can not parse value."$)
		Return 
	End Try

	'Create 2 rect for the object base and the chart
	
	'Create rectbase
	Dim objbase As TPFDObject
	objbase.Initialize
	objbase.width = obj.width
	objbase.height = obj.height
	objbase.centerx = obj.centerx
	objbase.centery = obj.centery
	objbase.x1 = obj.centerx - (objbase.width/2)
	objbase.y1 = obj.centery - (objbase.height/2)
	objbase.x2 = obj.centerx + (objbase.width/2)
	objbase.y2 = obj.centery + (objbase.height/2)
	objbase.fillcolor = obj.fillcolor
	objbase.bordercolor = obj.bordercolor
	objbase.textcolor = obj.textcolor
	objbase.strokewidth = obj.strokewidth
	
	Dim rectbase As B4XRect
	rectbase.Initialize(objbase.x1,objbase.y1,objbase.x2,objbase.y2)
	'Draw the inner shape
	cvs.DrawRect(rectbase, objbase.fillcolor, True, objbase.strokewidth)
	'Draw the outer shape
	If objbase.strokewidth > 0 Then cvs.DrawRect(rectbase, objbase.bordercolor, False, objbase.strokewidth)
	'Draw the title
	cvs.DrawText(title, objbase.centerx, obj.centery - (objbase.height/2) + fnt.size, fnt, obj.textcolor, "CENTER")

	'Create rectchart within the rectbase
	Dim objchart As TPFDObject
	objchart.Initialize
	objchart.width = obj.width - (chartoffset * 2)
	objchart.height = obj.height - (chartoffset * 2) - objchart.strokewidth
	objchart.x1 = obj.centerx - (objchart.width/2)
	objchart.y1 = obj.centery - (objchart.height/2) + objchart.strokewidth - (chartoffset / 2)
	objchart.x2 = obj.centerx + (objchart.width/2)
	objchart.y2 = obj.centery + (objchart.height/2) '- (chartoffset / 2)
	objchart.fillcolor = obj.fillcolor
	objchart.bordercolor = obj.bordercolor
	objchart.textcolor = obj.textcolor
	objchart.strokewidth = obj.strokewidth
	Dim rectchart As B4XRect
	rectchart.Initialize(objchart.x1,objchart.y1,objchart.x2,objchart.y2)
	'Draw the inner shape
	cvs.DrawRect(rectchart, objchart.fillcolor, True, objchart.strokewidth)
	'Draw the outer shape
	If objbase.strokewidth > 0 Then cvs.DrawRect(rectchart, objchart.bordercolor, False, objchart.strokewidth)
	
	'Draw y-scale bottom-up
	Dim dy As Float = (objchart.height / (scaleyticks))
	Dim dscaley As Int = (scaleymax - scaleymin) / scaleyticks
	'Log($"objchart y-scale: ${objchart.x1},${objchart.y1},${objchart.x2},${objchart.y2},${objchart.width},${objchart.height},${dy}"$)
	For i = 0 To scaleyticks
		x1 = objchart.x1 - scaleticksize
		x2 = objchart.x1
		y1 = objchart.y1 + (i * dy)
		y2 = y1
		cvs.DrawLine(x1,y1,x2,y2,objchart.bordercolor,objchart.strokewidth)
		cvs.DrawText( (scaleyticks - i) * dscaley, objchart.x1 - scaleticksize - 1, y1 + (fnt.Size/2), fnt, obj.textcolor, "RIGHT")
		'Log($"x1=${x1},y1=${y1},x2=${x2},y2=${y2},dy=${dy}"$)
	Next
	
	'Draw x-scale left-right
	'The first tick is a x=0 and the number of ticks datapoint size. The scale segments are datapoint size - 1.
	Dim dx As Float = ( (objchart.width) / (data.size - 1) )
	'Log($"objchart x-scale: datapoints=${data.size}, width=${objchart.width}, dx=${dx}, ${data.Get(0).As(Map).get("x")}-${data.Get(data.Size-1).as(Map).get("x")}"$)
	'Log($"objchart x-scale: ${objchart.x1},${objchart.y1},${objchart.x2},${objchart.y2},${objchart.width},${objchart.height},${dx}"$)
	For i = 0 To scalexticks - 1
		x1 = objchart.x1 + (i * dx) ' - (objchart.strokewidth / 2)
		x2 = x1
		y1 = objchart.y2
		y2 = objchart.y2 + scaleticksize
		cvs.DrawLine(x1,y1,x2,y2,objchart.bordercolor,objchart.strokewidth)
		'Log($"x1=${x1},y1=${y1},x2=${x2},y2=${y2},w=${objchart.width},h=${objchart.height},dx=${dx}"$)
	Next
	
	'Draw points
	'Set the base x,y at chart 0,0
	x1 = objchart.x1
	y1 = Helper.MapRangeFloat(0, scaleymin, scaleymax, objchart.y2, objchart.y1)
	For i = 0 To scalexticks - 1
		Dim p As Map = data.Get(i)
		x2 = objchart.x1 + (i*dx)
		y2 = Helper.MapRangeFloat(p.Get("y"), scaleymin, scaleymax, objchart.y2, objchart.y1)
		'Draw point indicator
		cvs.DrawPixel2(x2,y2,obj.bordercolor,1)
		'Draw line from previous to current point
		cvs.DrawLine(x1,y1,x2,y2,obj.bordercolor,obj.strokewidth/2)
		x1 = x2
		y1 = y2
		'Draw scale x-axis
		If scalextextrotation == 0 Then
			cvs.DrawTextRotated(p.Get("x"), _
							objchart.x1 + (i * dx), _ 
							objchart.y2 + (fnt.size), _ 
							fnt, objchart.textcolor, "CENTER", scalextextrotation)
		Else
			cvs.DrawTextRotated(p.Get("x"), _
							objchart.x1 + (i * dx) - (fnt.size / 2), _ 
							objchart.y2 + scaleticksize + 2, _ 
							fnt, objchart.textcolor, "LEFT", scalextextrotation)
		End If
	Next
End Sub
#End Region

#Region Valve
'Draw a 2-way-valve which has 2 triangles left+right or up+down and a dot in the middle.
'Object Properties:
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: VLV
Public Sub DrawValve2Way(cvs As AsyncCanvas, obj As TPFDObject)
	Dim objva As TPFDObject

	objva.Initialize
	objva.bordercolor = obj.bordercolor
	objva.fillcolor = obj.fillcolor
	objva.strokewidth = obj.strokewidth

	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If

	Select obj.direction
		Case PFDCommon.DIRECTION_HORIZONTAL
			'Draw the right arrow
			objva.centerx = obj.centerx - (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height
			objva.direction = PFDCommon.DIRECTION_RIGHT
			PFDShapesBasic.DrawArrow(cvs, objva)

			'Draw the left arrow
			objva.centerx = obj.centerx + (obj.width/4)
			objva.centery = obj.centery
			objva.direction = PFDCommon.DIRECTION_LEFT
			PFDShapesBasic.DrawArrow(cvs, objva)

		Case PFDCommon.DIRECTION_VERTICAL
			'Draw the down arrow
			objva.centerx = obj.centerx
			objva.centery = obj.centery - (obj.height/4)
			objva.width = obj.width
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_DOWN
			PFDShapesBasic.DrawArrow(cvs, objva)

			'Draw the up arrow
			objva.centerx = obj.centerx
			objva.centery = obj.centery + (obj.height/4)
			objva.direction = PFDCommon.DIRECTION_UP
			PFDShapesBasic.DrawArrow(cvs, objva)
	End Select

	'Draw a dot at valve centerxy pos
	cvs.DrawCircle(obj.centerx, obj.centery, obj.strokewidth * 2, obj.bordercolor, True, 0)

	'Draw the tag & value
	PFDCommon.DrawTagValue(cvs, obj)

	'Example using drawcheck
	'PFDShapesBasic.DrawCheck(cvs, obj)
	
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw a 2-way-valve (same as DrawValve2Way).
'Valve state: 0=closed,100=open,9999=error
'x,y (int) - TopLeftposition of the object.
'PID Abbreviation: VLV
Public Sub DrawValve(cvs As AsyncCanvas, obj As TPFDObject)
	DrawValve2Way(cvs, obj)
End Sub

'Draw a 3-way-valve which has 3 arrows left+right+up/down for horizontal or other way around for vertical and a dot in the middle.
'Direction Vertical recommended width 24, height 40.
'Object Properties:
'centerx,centery,width,height,direction,tag,value,unit,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: VLV
Public Sub DrawValve3Way(cvs As AsyncCanvas, obj As TPFDObject)
	Dim objva As TPFDObject
	Dim xvu, yvu As Int

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)
	Dim fntrect As B4XRect = cvs.MeasureText(obj.value, fnt)

	objva.Initialize
	objva.bordercolor = obj.bordercolor
	objva.fillcolor = obj.fillcolor
	objva.strokewidth = obj.strokewidth

	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL_UP
	End If

	Select obj.direction
		Case PFDCommon.DIRECTION_HORIZONTAL_UP
			objva.centerx = obj.centerx - (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_RIGHT
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx + (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_LEFT
			PFDShapesBasic.DrawArrow(cvs, objva)

			objva.centerx = obj.centerx
			objva.centery = obj.centery - (obj.height/4) 	'- (obj.strokewidth * 2)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_DOWN
			PFDShapesBasic.DrawArrow(cvs, objva)

			xvu = obj.centerx
			yvu = obj.centery + (obj.height/2)
			
		Case PFDCommon.DIRECTION_HORIZONTAL_DOWN
			objva.centerx = obj.centerx - (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_RIGHT
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx + (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_LEFT
			PFDShapesBasic.DrawArrow(cvs, objva)

			objva.centerx = obj.centerx
			objva.centery = obj.centery + (obj.height/4) 	'- (obj.strokewidth * 2)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_UP
			PFDShapesBasic.DrawArrow(cvs, objva)

			xvu = obj.centerx
			yvu = obj.centery - (obj.height/4)

		Case PFDCommon.DIRECTION_VERTICAL_RIGHT
			objva.centerx = obj.centerx
			objva.centery = obj.centery - (obj.height/4)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_DOWN
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx
			objva.centery = obj.centery + (obj.height/4)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_UP
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx + (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_LEFT
			PFDShapesBasic.DrawArrow(cvs, objva)

			xvu = obj.centerx - (obj.width/4) - (fntrect.Width/4)
			yvu = obj.centery + (fntrect.Height/2)

		Case PFDCommon.DIRECTION_VERTICAL_LEFT
			objva.centerx = obj.centerx
			objva.centery = obj.centery - (obj.height/4)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_DOWN
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx
			objva.centery = obj.centery + (obj.height/4)
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_UP
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			objva.centerx = obj.centerx - (obj.width/4)
			objva.centery = obj.centery
			objva.width = obj.width/2
			objva.height = obj.height/2
			objva.direction = PFDCommon.DIRECTION_RIGHT
			PFDShapesBasic.DrawArrow(cvs, objva)
			
			xvu = obj.centerx + (obj.width/4) + (fntrect.Width/4)
			yvu = obj.centery + (fntrect.Height/2)
	End Select

	'Draw a dot at valve centerxy pos
	cvs.DrawCircle(obj.centerx, obj.centery, obj.strokewidth * 2, obj.bordercolor, True, 0)

	'Draw the value and unit if length > 0
	If obj.value.Length > 0 Then
		cvs.DrawText($"${obj.value}${obj.unit}"$, xvu, yvu, fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region Vessel
'Draw Vessel.
'Object Properties:
'centerx,centery,width,height,direction,tag,bordercolor,fillcolor,textcolor,textalignment.
'PID Abbreviation: V
Public Sub DrawVessel(cvs As AsyncCanvas, obj As TPFDObject)
	Dim x1,y1,x2,y2 As Int
	Dim p As B4XPath
	Dim rect As B4XRect

	'Set the font
	Dim fnt As B4XFont = PFDCommon.SetFont(obj)

	'Check text alignment
	If obj.textalignment == PFDCommon.TEXT_ALIGNMENT_NONE Then obj.textalignment = PFDCommon.TEXT_ALIGNMENT_CENTER
	
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rect.Initialize(x1,y1,x2,y2)

	If obj.direction == PFDCommon.DIRECTION_NONE Then
		obj.direction = PFDCommon.DIRECTION_HORIZONTAL
	End If

	If obj.direction == PFDCommon.DIRECTION_HORIZONTAL Then
		p.InitializeRoundedRect(rect, obj.height/2)
	End If
	If obj.direction == PFDCommon.DIRECTION_VERTICAL Then
		p.InitializeRoundedRect(rect, obj.width/2)
	End If

	'Draw the inner shape
	cvs.DrawPath(p, obj.fillcolor, True, 0)
	'Draw the outer shape
	cvs.DrawPath(p, obj.bordercolor, False, obj.strokewidth)

	'Show the tag in the object middle
	If obj.tag.Length > 0 Then
		Dim recttag As B4XRect = cvs.MeasureText(obj.tag, fnt)
		Dim oxtag As Int = obj.centerx
		Dim oytag As Int = obj.centery + (recttag.height/2)
		Dim strtag As String = obj.tag
		If recttag.Width > obj.width Then
			strtag = strtag.Replace(" ", CRLF)
			oxtag = obj.centerx
			oytag = obj.centery - (recttag.height/2)
		End If
		cvs.DrawText(strtag, oxtag, oytag, fnt, obj.textcolor, obj.textalignment)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region ISA101

#Region BinaryStateIndicator
'Draw binary state indicator according ISA-101.
'The shape has a rectangle showing 
'state true or false: true=filled, false=empty
'text: text which remains the same, i.e. does not change if state changes
'Properties: description color=darkgray
Public Sub DrawBinaryStateIndicator(cvs As AsyncCanvas, obj As TPFDObject)
	'Set the ISA101 properties
	Dim state_rect_strokewidth As Int = 1
	Dim state_rect_color_true As Int = xui.Color_RGB(192,192,192)
	Dim state_rect_color_false As Int = xui.color_white
	Dim state_fontsize As Int = obj.fontsize
	Dim state_fontcolor As Int = xui.Color_DarkGray

	'Rect shape
	Dim rectshape As B4XRect
	Dim x1,y1,x2,y2 As Int
	'Center pos of the text
	Dim cx As Float
	Dim cy As Float
	
	'Rect state
	Dim rectstate As B4XRect
	'Rect filled color depending state (obj.value) is true (1) or false (0)
	Dim rectstatecolor As Int
	'Rect dimension
	Dim rsx1,rsy1,rsx2,rsy2 As Int
	'Rect x offset (left), y is centered
	Dim rsxoffset As Int = 5
	'Rect fixed size 10x10
	Dim rsw As Int = 10
	Dim rsh As Int = 10

	'Check if shapeselector same as pfdobject which means the shape is drawn for the shapeselector
	'Reduce the fontsize and the text offset from the rect state
	If obj.shapeselector == obj.pfdobject Then
		rsxoffset = 2
		state_fontsize = 6
	End If

	'Rect shape
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rectshape.Initialize(x1,y1,x2,y2)
	If obj.strokewidth > 0 Then
		cvs.DrawRect(rectshape, obj.bordercolor, False, obj.strokewidth)		
	End If

	'Rect state
	rsx1 = obj.centerx - (obj.width/2) + rsxoffset
	rsy1 = obj.centery - (rsh/2)
	rsx2 = rsx1 + rsw
	rsy2 = obj.centery + (rsh/2)
	rectstate.Initialize(rsx1,rsy1,rsx2,rsy2)
	'Set rect state filled color depending state 0=False, 1=True
	rectstatecolor = IIf(obj.value == 0, state_rect_color_false, state_rect_color_true)
	cvs.DrawRect(rectstate, rectstatecolor, True, state_rect_strokewidth)

	'Draw the text left justified at right side of the rect state
	Dim fnt As B4XFont = xui.CreateDefaultFont(state_fontsize)
	Dim fntvaluerect As B4XRect = cvs.MeasureText("A", fnt)
	'Set the description right from the rect state
	cx = rsx2 + (rsxoffset/2)
	cy = obj.centery + (fntvaluerect.Height / 2)
	cvs.DrawText(obj.description, cx, cy, fnt, state_fontcolor, PFDCommon.TEXT_ALIGNMENT_LEFT)
	
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region NumericData
'Draw numeric data according ISA-101.
'The shape has 2 horizontal lines: Top=Description, Bottom=Value with Unit
'Properties: description color=darkgray,value color=blue,unit color=darkgray.
Public Sub DrawNumericData(cvs As AsyncCanvas, obj As TPFDObject)
	'Center pos of the shape
	Dim cx As Float = obj.centerx
	Dim cy As Float = obj.centery - 4
	
	'Shape Height - if shape for shape selector then change to fix value
	'Dim h As Int = obj.height
	
	'Fontsize - if shape for shape selector then change to very small
	Dim fs As Int = obj.fontsize
	
	'Value color - if shape for shape selector then change to darkgray
	Dim valuecolor As Int = obj.textcolor	'xui.Color_Blue
	
	'Special case set fontsize for the shape in the shapeselector
	If obj.shapeselector == obj.value Then
		fs = 8	'fs - 6
		valuecolor = xui.Color_DarkGray
	End If

	'Top line: description
	Dim fnt As B4XFont = xui.CreateDefaultFont(fs)
	cvs.DrawText(obj.description, cx, cy, fnt, xui.Color_DarkGray, "CENTER")
	
	'Bottom line: value
	If obj.value.Length > 0 Then
		Dim fntvalue As B4XFont = PFDCommon.SetFont(obj)	'xui.CreateDefaultBoldFont(fs + 2)
		Dim fntvaluerect As B4XRect = cvs.MeasureText(obj.value, fnt)
		cx = obj.centerx
		cy = obj.centery + (fntvaluerect.Height * 1.5)
		cvs.DrawText(obj.value, cx, cy, fntvalue, valuecolor, PFDCommon.TEXT_ALIGNMENT_CENTER)
	End If

	'Bottom line: unit
	If obj.unit.Length > 0 Then
		cx = obj.centerx + (fntvaluerect.width / 2) + 4
		cy = obj.centery + (fntvaluerect.Height * 1.5)
		cvs.DrawText(obj.unit, cx, cy, fnt, xui.Color_DarkGray, PFDCommon.TEXT_ALIGNMENT_LEFT)
	End If

	Dim rect As B4XRect
	Dim x1,y1,x2,y2 As Int
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	rect.Initialize(x1,y1,x2,y2)

	'NOT USED = Draw the inner shape
	'cvs.DrawRect(rect, obj.fillcolor, True, 0)
	'Draw the outer shape
	If obj.strokewidth > 0 Then
		cvs.DrawRect(rect, obj.bordercolor, False, obj.strokewidth)
	End If

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region PrimaryProcessLine
'Draw a primary process line according ISA-101.
'Properties: strokewidth=3,fillcolor=Gray 160 (#A0A0A4).
Public Sub DrawPrimaryProcessLine(cvs As AsyncCanvas, obj As TPFDObject)
	'Set the ISA101 properties
	'obj.fillcolor = xui.Color_RGB(160,160,160)
	'obj.strokewidth = 3
	
	'Draw the line
	PFDShapesBasic.DrawLineSolid(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region SecondaryProcessLine
'Draw a Secondary process line according ISA-101.
'Properties: strokewidth=1,fillcolor=Gray 160 (#A0A0A4).
Public Sub DrawSecondaryProcessLine(cvs As AsyncCanvas, obj As TPFDObject)
	'Set the ISA101 properties
	'obj.fillcolor = xui.Color_RGB(160,160,160)
	'obj.strokewidth = 1
	
	'Draw the line
	PFDShapesBasic.DrawLineSolid(cvs, obj)

	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#Region DataConnectorLine
'Draw a data connector line according ISA-101.
'Properties: dashed, strokewidth=1,fillcolor=Gray 160 (#A0A0A4).
Public Sub DrawDataConnectorLine(cvs As AsyncCanvas, obj As TPFDObject)
	Dim dclobj As TPFDObject
	
	'Init the object
	dclobj.Initialize
	dclobj.shapeselector = obj.shapeselector
	
	'Set the ISA101 properties
	dclobj.fillcolor	= obj.fillcolor		
	dclobj.strokewidth	= obj.strokewidth	

	'Set pos,w,h,direction
	dclobj.centerx		= obj.centerx
	dclobj.centery		= obj.centery
	dclobj.width		= obj.width
	dclobj.height		= obj.height
	dclobj.direction	= obj.direction
	If obj.direction == PFDCommon.DIRECTION_NONE Then
		dclobj.direction = PFDCommon.DIRECTION_HORIZONTAL_DOWN
	End If

	'Adjust obj w&h to fit within shape selector
	If obj.shapeselector == obj.pfdobject Then
		dclobj.width = PFDCommon.SHAPE_SELECTOR_WIDTH * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
		dclobj.height = PFDCommon.SHAPE_SELECTOR_HEIGHT * PFDCommon.SHAPE_SELECTOR_RECT_FACTOR
	End If

	'Set obj cx,cy,fillcolor,strokewidth
	Dim cx As Int = dclobj.centerx
	Dim cy As Int = dclobj.centery

	'Direction UP with 2 lines hor + vert
	'-+
	' |
	If dclobj.direction == PFDCommon.DIRECTION_UP Then
		'TopLeft to TopRight = hor
		dclobj.centerx = cx
		dclobj.centery = cy - (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_HORIZONTAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)

		'TopRight to BotRight = vert
		dclobj.centerx = cx + (dclobj.width/2)
		dclobj.centery = cy '+ (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_VERTICAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)
	End If

	'Direction LEFT with 2 lines hor + vert
	' |
	'-+
	If dclobj.direction == PFDCommon.DIRECTION_LEFT Then
		'BotLeft to BotRight = hor
		dclobj.centerx = cx
		dclobj.centery = cy + (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_HORIZONTAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)

		'BotRight to TopRight = vert
		dclobj.centerx = cx + (dclobj.width/2)
		dclobj.centery = cy '+ (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_VERTICAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)
	End If

	'Direction RIGHT with 2 lines vert +  hor
	'|
	'+-
	If dclobj.direction == PFDCommon.DIRECTION_RIGHT Then
		'TopLeft to BotLeft = vert
		dclobj.centerx = cx - (dclobj.width/2)
		dclobj.centery = cy
		dclobj.direction = PFDCommon.DIRECTION_VERTICAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)

		'BotLeft to BotRight = hor
		dclobj.centerx = cx
		dclobj.centery = cy + (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_HORIZONTAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)
	End If
	
	'Direction DOWN with 2 lines vert + hor
	'+-
	'|
	If dclobj.direction == PFDCommon.DIRECTION_DOWN Then
		'TopLeft to TopRight = hor
		dclobj.centerx = cx
		dclobj.centery = cy - (dclobj.height/2)
		dclobj.direction = PFDCommon.DIRECTION_HORIZONTAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)

		'TopLeft to BotLeft = vert
		dclobj.centerx = cx - (dclobj.width/2)
		dclobj.centery = cy
		dclobj.direction = PFDCommon.DIRECTION_VERTICAL
		PFDShapesBasic.DrawLineDashed(cvs,dclobj)
	End If
	
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub

'Draw a data connector line according ISA-101.
'Properties: strokewidth=1,fillcolor=Gray 160 (#A0A0A4).
Public Sub DrawConnectorLine4(cvs As AsyncCanvas, obj As TPFDObject)
	'Set the ISA101 properties
	Dim fillcolor As Int = xui.Color_Red'xui.Color_RGB(160,160,160)
	Dim strokewidth As Int = 4
	
	'Flag to set number of lines
	Dim lines As Int = 2

	'Rect shape
	Dim x1,y1,x2,y2 As Int
	Dim w2,h2 As Int
	
	'Center pos of the text
	'Dim cx As Float
	'Dim cy As Float

	'Get the dimension of the line start and end as rect
	x1 = obj.centerx - (obj.width/2)
	y1 = obj.centery - (obj.height/2)
	x2 = obj.centerx + (obj.width/2)
	y2 = obj.centery + (obj.height/2)
	w2 = Abs(x2-x1)/2
	h2 = Abs(y2-y1)/2
	
	'TESTS
	'rectshape.Initialize(x1,y1,x2,y2)
	'cvs.DrawRect(rectshape, obj.bordercolor, False, 0.5)

	'TEST start topleft to bottom right with 2 lines
	If lines == 2 Then
		'TopLeft to TopRight
		x1 = obj.centerx - (obj.width/2)
		y1 = obj.centery - (obj.height/2)
		x2 = obj.centerx + (obj.width/2)
		y2 = obj.centery - (obj.height/2)
		'Log($"Line1=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)

		'TopRight to BotRight
		x1 = obj.centerx + (obj.width/2)
		y1 = obj.centery - (obj.height/2)
		x2 = obj.centerx + (obj.width/2)
		y2 = obj.centery + (obj.height/2)
		'Log($"Line2=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)
	End If
	
	If lines == 4 Then
		'TopLeft to TopMid
		x1 = obj.centerx - (obj.width/2)
		y1 = obj.centery - (obj.height/2)
		x2 = x1 + w2
		y2 = y1
		'Log($"Line1=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)

		'TopMid to MidMid (center)
		x1 = x2
		y1 = y2
		x2 = x2
		y2 = y1 + h2
		'Log($"Line1=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)

		'MidMid to MidRight
		x1 = x2
		y1 = y2
		x2 = x1 + w2
		y2 = y1
		'Log($"Line2=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)

		'MidRight to BotRight
		x1 = x2
		y1 = y2
		x2 = x2
		y2 = y1 + h2
		'Log($"Line2=${x1},${y1},${x2},${y2},"$)
		cvs.DrawLine(x1,y1,x2,y2,fillcolor,strokewidth)
	End If
	
	'Commit the drawing is done in method ReDraw in PFDBase.
	'cvs.Invalidate
End Sub
#End Region

#End Region
